import base64
from typing import Annotated, Any

import numpy as np
from pydantic import PlainSerializer, PlainValidator, WithJsonSchema
from pydantic_core import core_schema


def validate_value(value: Any) -> Any:
    """Validate (i.e. deserialize) JSON serializable value to Python type, to support complex and ndarray types."""
    if isinstance(value, dict):
        if "__complex__" in value:
            value = complex(value["real"], value["imag"])
        elif "__ndarray__" in value:
            data = base64.b64decode(value["data"])
            value = np.frombuffer(data, value["dtype"]).reshape(value["shape"])
    return value


def serialize_value(value: Any) -> Any:
    """Serialize value type to JSON serializable type, to support complex and ndarray types."""
    if isinstance(value, complex):
        value = {"__complex__": "true", "real": value.real, "imag": value.imag}
    elif isinstance(value, np.ndarray):
        # ensure array buffer is contiguous and in C order
        value = np.require(value, requirements=["A", "C"])
        data = base64.b64encode(value.data)
        value = {"__ndarray__": "true", "data": data, "dtype": str(value.dtype), "shape": value.shape}
    return value


# TODO: We might want to rename these to ObservationValue and ObservationUncertainty, respectively.
Value = Annotated[
    bool | str | int | float | complex | np.ndarray,
    PlainValidator(validate_value),
    PlainSerializer(serialize_value),
    WithJsonSchema(core_schema.any_schema()),
]
Uncertainty = Annotated[
    int | float | complex | np.ndarray,
    PlainValidator(validate_value),
    PlainSerializer(serialize_value),
    WithJsonSchema(core_schema.any_schema()),
]

# TODO (Marko): Consider if we want to rename these?
#  We have "as" imports so seems like more descriptive name would be better.
ObservationValue = Value
ObservationUncertainty = Uncertainty



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convert SettingNodes to protos and back."""

import iqm.data_definitions.common.v1.setting_pb2 as spb

from exa.common.api.proto_serialization import datum
import exa.common.api.proto_serialization._parameter as param_proto
from exa.common.data.parameter import DataType, Parameter
from exa.common.data.setting_node import Setting, SettingNode
from exa.common.helpers.numpy_helper import coerce_numpy_type_to_native


def _pack_setting(setting: Setting, optimize: bool) -> spb.SettingNode.Setting:
    """Convert a Setting into protobuf representation."""
    value = coerce_numpy_type_to_native(setting.value)
    try:
        packed = datum.pack(value)
    except ValueError as err:
        raise ValueError(
            f"Failed to convert a value to protobuf. Value={setting.value}, Parameter={setting.parameter}."
        ) from err
    if optimize:
        return spb.SettingNode.Setting(parameter_name=setting.name, value=packed)
    return spb.SettingNode.Setting(parameter=param_proto.pack(setting.parameter), value=packed)


def _unpack_setting(proto: spb.SettingNode.Setting) -> Setting:
    """Convert protobuf representation into a Setting."""
    if proto.WhichOneof("parameter_desc") == "parameter":
        parameter = param_proto.unpack(proto.parameter)
    else:
        parameter = Parameter(name=proto.parameter_name, data_type=DataType.ANYTHING)
    try:
        value = datum.unpack(proto.value)
    except Exception as err:
        raise AttributeError(f"Unpacking of {parameter} {proto.value} failed.") from err

    return Setting(parameter, value)


def pack(node: SettingNode, minimal: bool) -> spb.SettingNode:
    """Convert a SettingNode into protobuf representation.

    Silently coerces some datatypes to be compatible with the proto definition of ``Datum``:
    - Numpy arrays of 32-bit ints are converted to 64-bits (Windows only).
    - Singular numpy types are converted into corresponding native types.

    Args:
        node: SettingNode to pack, recursively.
        minimal: If True, only the :attr:`.Parameter.name` of each Setting is preserved along with the setting value.
            If False, the the whole :attr:`.Setting.parameter` is packed.

    Returns:
        Protobuf instance that represents `node`.

    """
    settings = {key: _pack_setting(item, minimal) for key, item in node.child_settings}
    nodes = {key: pack(item, minimal) for key, item in node.child_nodes}
    return spb.SettingNode(name=node.name, settings=settings, subnodes=nodes)


def unpack(proto: spb.SettingNode) -> SettingNode:
    """Convert protobuf representation into a SettingNode. Reverse operation of :func:`.pack`

    Args:
        proto: Protobuf instance to unpack, recursively.

    Returns:
        Unpacked SettingNode. In case `proto` only contains the parameter names (see ``optimize`` in
        :func:`.pack`), dummy Parameters are generated.

    """
    settings = {key: _unpack_setting(content) for key, content in proto.settings.items()}
    nodes = {key: unpack(content) for key, content in proto.subnodes.items()}
    # Names are currently NEVER aligned with the paths when deserializing. This is safe to do, since currently nothing
    # in the server-side assumes path==name, but if such logic is added this needs to be reconsidered.
    return SettingNode(name=proto.name, **(settings | nodes), align_name=False)



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sweep specification with linearly spaced values."""

from typing import Any
import warnings

from exa.common.control.sweep.option import CenterSpanOptions, StartStopOptions
from exa.common.control.sweep.sweep import Sweep
from exa.common.data.parameter import Parameter
from exa.common.errors.exa_error import InvalidSweepOptionsTypeError


class LinearSweep(Sweep):
    """Generates evenly spaced parameter values based on `options`.

    - If `options` is instance of :class:`.StartStopOptions`, then start and stop options are used for interval
    - If `options` is instance of :class:`.CenterSpanOptions`,
      then the start and stop of the interval are calculated from center and span values

    Raises:
        ValueError: Error is raised if `options` is inconsistent.

    """

    def __init__(
        self,
        parameter: Parameter,
        options: StartStopOptions | CenterSpanOptions | None = None,
        *,
        data: list[Any] | None = None,
        **kwargs,
    ) -> None:
        warnings.warn("LinearSweep is deprecated, use Sweep instead.", DeprecationWarning)
        if options and not isinstance(options, StartStopOptions | CenterSpanOptions):
            raise InvalidSweepOptionsTypeError(str(type(options)))
        super().__init__(parameter, options, data=data, **kwargs)



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper constants for SweepOptions classes."""

from typing import Any

#: Default value for `count` value in options.
DEFAULT_COUNT: int = 101
#: Default value for `base` value in options.
DEFAULT_BASE: int = 10
#: Dictionary with all possible types of options
OPTIONS_TYPE: dict[str, Any] = {
    "start": "start",
    "stop": "stop",
    "center": "center",
    "span": "span",
    "count": "count",
    "step": "step",
    "asc": "asc",
    "start_exp": "start_exp",
    "stop_exp": "stop_exp",
    "center_exp": "center_exp",
    "span_exp": "span_exp",
    "fixed": "fixed",
    "start_stop_list": ["start", "stop", "count", "step"],
    "center_span_list": ["center", "span", "count", "step", "asc"],
    "start_stop_base_list": ["start", "stop", "count", "base"],
    "center_span_base_list": ["center", "span", "count", "base", "asc"],
}



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Conversions between Protocol buffers in :mod:`iqm.data_structures.common` and Python objects in exa-common.

Each submodule corresponds to a Protobuf definition from :mod:`iqm.data_structures.common`.
Each module has at least 2 functions, ``pack`` and ``unpack``, which pack the Python type into a Protobuf instance or
vice versa.

The packed objects can be nested when constructing new Protobuf objects.
It is intended that the user does the final serialization using the Protobuf service methods.
"""

from iqm.data_definitions.common.v1.setting_pb2 import SettingNode as SettingNodeProto
from iqm.data_definitions.common.v1.sweep_pb2 import CartesianSweep as CartesianSweepProto

import exa.common.api.proto_serialization.array as array
import exa.common.api.proto_serialization.datum as datum
import exa.common.api.proto_serialization.nd_sweep as nd_sweep
import exa.common.api.proto_serialization.sequence as sequence
import exa.common.api.proto_serialization.setting_node as setting_node



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Range specification to define a range around a center value."""

from dataclasses import dataclass

from exa.common.control.sweep.option.constants import DEFAULT_COUNT
from exa.common.control.sweep.option.start_stop_options import StartStopOptions
from exa.common.control.sweep.option.sweep_options import SweepOptions


@dataclass(frozen=True)
class CenterSpanOptions(SweepOptions):
    """Range generation options.

    Values are generated over the interval with the center `center` and the size of `span`.
    For linear sweep range the number of generated values can be based either on `count` or `step`.
    In case `count` is empty and `step` is not, `step` is used for calculating `count`.
    For exponential sweep range only `count` is used.
    """

    #: Value of interval center.
    center: int | float | complex
    #: Size of the interval.
    span: int | float | complex
    #: Number of values to generate.
    #: If `count` and `step` are empty, the default value of count is
    #: :const:`exa.common.control.sweep.option.constants.DEFAULT_COUNT`.
    count: int | None = None
    #: Size of spacing between values.
    step: int | float | complex = None
    #: Order of generated values. Default to ascending
    asc: bool | None = None

    def __post_init__(self):
        if self.count is not None and self.step is not None:
            object.__setattr__(self, "step", None)
        if self.count is None and self.step is None:
            object.__setattr__(self, "count", DEFAULT_COUNT)
        if self.asc is None:
            object.__setattr__(self, "asc", True)

    @property
    def data(self) -> list[int | float | complex]:
        start = self.center - (self.span / 2)
        stop = self.center + (self.span / 2)
        (start, stop) = (start, stop) if self.asc else (stop, start)
        return StartStopOptions(start, stop, count=self.count, step=self.step).data



# -*- coding: utf-8 -*-
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

from packaging.version import parse

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
py_path = os.path.join(os.getcwd(), os.path.dirname(__file__), "../src")
sys.path.insert(0, os.path.abspath(py_path))


# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = "7.2"

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "exa-common"
copyright = "2019-2022, IQM Finland Oy"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
# The short X.Y version.
version = ""
# The full version, including alpha/beta/rc tags.
release = ""
try:
    from exa.common import __version__ as release
except ImportError:
    pass
else:
    version = parse(release).base_version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = "%Y-%m-%d"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "_templates"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# Shorten class names in function type hints
python_use_unqualified_type_names = True


# -- Autodoc ------------------------------------------------------------

# member ordering in autodoc output (default: 'alphabetical')
autodoc_member_order = "bysource"

# mock some imports when building the docs
autodoc_mock_imports = [
    "iqm.data_definitions.common.v1.data_types_pb2",
    "iqm.data_definitions.common.v1.setting_pb2",
    "iqm.data_definitions.common.v1.sweep_pb2",
]

# where should signature annotations appear in the docs, function signature or parameter description?
autodoc_typehints = "description"
# autodoc_typehints = 'description' puts the __init__ annotations into its docstring,
# which we thus have to include in the class documentation.
autoclass_content = "both"

# Sphinx 3.3+: manually clean up type alias rendering in the docs
# autodoc_type_aliases = {'TypeAlias': 'exa.experiment.somemodule.TypeAlias'}


# -- Autosummary ------------------------------------------------------------

# use autosummary to generate stub pages for API docs
autosummary_generate = True


# -- Options for HTML output ---------------------------------------------------

import sphinx_book_theme

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_book_theme"

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [sphinx_book_theme.get_html_theme_path()]

html_context = dict(display_github=False)

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "collapse_navigation": True,
    "navigation_depth": 5,
    "use_download_button": False,
}

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/images/logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/images/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%Y-%m-%d"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False
html_copy_source = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "exa-common-doc"


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    "papersize": "a4paper",
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ("index", "user_guide.tex", "exa-common Documentation", "IQM Finland Oy", "manual"),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = ""

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- External mapping ------------------------------------------------------------

use_local_target = os.getenv("USE_LOCAL_TARGET", "").lower()

python_version = ".".join(map(str, sys.version_info[0:2]))
intersphinx_mapping = {
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
    "python": ("https://docs.python.org/" + python_version, None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "xarray": ("https://docs.xarray.dev/en/stable", None),
}

if use_local_target != "true":
    extlinks = {
        "issue": ("https://jira.iqm.fi/browse/%s", "issue %s"),
        "mr": ("https://gitlab.iqm.fi/iqm/qccsw/exa/exa-repo/-/merge_requests/%s", "MR %s"),
    }



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Chip topology class for parsing CHAD and other QPU related data into human-usable form."""

from __future__ import annotations

from collections.abc import Collection, Iterable
import itertools
import re
from typing import TypeVar

from exa.common.qcm_data.chad_model import CHAD

Locus = tuple[str, ...] | frozenset[str]

DEFAULT_1QB_MAPPING = "qubits"
DEFAULT_2QB_MAPPING = "connected_qubits"

ComponentMap = TypeVar("ComponentMap", bound=dict[str, Collection[str]])


def _get_numeric_id(name: str) -> int:
    """Sorting key for component names."""
    match = re.search(r"(\d+)", name)
    return int(match.group(1)) if match else 0


def _get_coupler_numeric_ids(name: str) -> tuple[int, ...]:
    """Sorting key for coupler names.

    Supports e.g. the following patterns: "TC-1-2", "TC1"
    """
    return tuple(int(m) for m in re.findall(r"\d+", name))


def sort_components(components: Iterable[str]) -> list[str]:
    """Sort the given components in a human-readable way."""
    return sorted(components, key=lambda x: (re.sub(r"[^a-zA-Z]", "", x), -_get_numeric_id(x), x), reverse=True)


def sort_couplers(couplers: Iterable[str]) -> list[str]:
    """Sort the given couplers in a human-readable way."""
    return sorted(couplers, key=_get_coupler_numeric_ids)


class ChipTopology:
    """Topology information for a chip (typically a QPU).

    Can represent the information found in a CHAD, as well as locus mappings for gates.

    Args:
        qubits: names of the qubits.
        computational_resonators: names of the computational resonators.
        couplers: mapping from coupler name to names of chip components it connects to.
        probe_lines: mapping from probe line name to names of chip components it connects to.
        variant: identifier of the QPU design variant.

    """

    def __init__(
        self,
        qubits: Iterable[str],
        computational_resonators: Iterable[str],
        couplers: dict[str, Iterable[str]],
        probe_lines: dict[str, Iterable[str]],
        variant: str = "",
    ):
        self.variant = variant
        self.qubits = frozenset(qubits)
        self.qubits_sorted = tuple(sort_components(self.qubits))

        self.computational_resonators = frozenset(computational_resonators)
        """Computational resonators on the chip, in any order."""
        self.computational_resonators_sorted = tuple(sort_components(self.computational_resonators))
        """Computational resonators on the chip, sorted."""

        data_components = self.qubits | self.computational_resonators
        if diff := set(itertools.chain.from_iterable(couplers.values())) - data_components:
            raise ValueError(f"Couplers connect to unknown components: {diff}")
        if diff := set(itertools.chain.from_iterable(probe_lines.values())) - (data_components | frozenset(couplers)):
            raise ValueError(f"Probe lines connect to unknown components: {diff}")

        self.couplers: frozenset[str] = frozenset(couplers)
        """Tunable couplers on the chip, in any order."""
        self.couplers_sorted: tuple[str, ...] = tuple(sort_couplers(self.couplers))
        """Tunable couplers on the chip, sorted by numerical IDs."""

        self.probe_lines: frozenset[str] = frozenset(probe_lines)
        """Probe lines on the chip, in any order."""
        self.probe_lines_sorted: tuple[str, ...] = tuple(sorted(self.probe_lines))
        """Probe lines on the chip, sorted."""

        self.all_components = data_components | self.couplers | self.probe_lines
        """All components on the chip."""

        self.coupler_to_components: dict[str, tuple[str, ...]] = {
            coupler: tuple(sort_components(components)) for coupler, components in couplers.items()
        }
        """Map from each coupler to all other components it connects to. The values are sorted."""
        component_to_couplers = {}
        for coupler, components in couplers.items():
            for c in components:
                component_to_couplers.setdefault(c, set()).add(coupler)
        self.component_to_couplers: dict[str, frozenset] = {
            c: frozenset(couplers) for c, couplers in component_to_couplers.items()
        }
        """Map from each component to all couplers connected to it."""
        self.probe_line_to_components: dict[str, tuple[str, ...]] = {
            pl: tuple(sort_components(components)) for pl, components in probe_lines.items()
        }
        """Map from each probe line to all components it connects to."""
        # NOTE: assumes just one pl per component
        self.component_to_probe_line = {c: pl for pl, components in probe_lines.items() for c in components}
        """Map from each component to the probeline connected to it.
        Max 1 connection per component is assumed.
        Components without connection to a probe line don't appear.
        """

        self._locus_mappings: dict[Locus, dict[tuple[str, ...]]] = {
            DEFAULT_1QB_MAPPING: {(qubit,): (qubit,) for qubit in self.qubits_sorted},
            DEFAULT_2QB_MAPPING: {
                frozenset(comps): (coupler,)
                for coupler, comps in self.coupler_to_components.items()
                if len(comps) == 2
                and set(comps).issubset(frozenset().union(*[self.qubits, self.computational_resonators]))
            },
        }

    @classmethod
    def from_chip_design_record(cls, record: dict) -> ChipTopology:
        """Construct a ChipTopology instance from a raw Chip design record.

        Args:
            record: Record as returned by Station control.

        Returns:
            Corresponding chip topology

        """
        return cls.from_chad(CHAD(**record))

    @classmethod
    def from_chad(cls, chad: CHAD) -> ChipTopology:
        """Construct a ChipTopology instance from a CHAD. Use :meth:`from_chip_design_record` if possible.

        Args:
            chad: parsed CHAD model
        Returns:
            corresponding chip topology

        """
        qubits = chad.qubit_names
        computational_resonators = chad.computational_resonator_names
        data_components = frozenset(qubits + computational_resonators)
        return cls(
            qubits=qubits,
            computational_resonators=computational_resonators,
            couplers={coupler.name: data_components & set(coupler.connections) for coupler in chad.components.couplers},
            probe_lines={
                pl.name: (data_components | frozenset(chad.coupler_names)) & set(pl.connections)
                for pl in chad.components.probe_lines
            },
            variant=chad.variant,
        )

    def get_neighbor_couplers(self, components: Iterable[str]) -> set[str]:
        """Couplers that connect to at least one of the given chip components.

        Args:
            components: some chip components, typically qubits and computational resonators
        Returns:
            couplers that connect to at least one of ``components``

        """
        couplers: set[str] = set()
        for component in components:
            if (coupler := self.component_to_couplers.get(component)) is not None:
                couplers |= coupler
        return couplers

    def get_connecting_couplers(self, components: Collection[str]) -> set[str]:
        """Couplers that only connect to the given chip components, and connect at least two of them.

        Equivalent to returning the edges in the ``components``-induced
        subgraph of the coupling topology.

        Args:
            components: some chip components, typically qubits and computational resonators
        Returns:
            couplers that connect to only members of ``components``, and to at least two of them

        """
        connecting_couplers = set()
        for coupler in self.get_neighbor_couplers(components):
            connections = self.coupler_to_components[coupler]
            if all(q in components for q in connections) and len(connections) >= 2:
                connecting_couplers.add(coupler)
        return connecting_couplers

    def get_coupler_for(self, component_1: str, component_2: str) -> str:
        """Common coupler for the given chip components (e.g. qubit or computational resonator).

        Args:
            component_1: first component
            component_2: second component
        Returns:
            the common coupler
        Raises:
            ValueError: the given components have zero or more than one connecting coupler

        """
        connecting_couplers = self.get_connecting_couplers((component_1, component_2))
        if (n_couplers := len(connecting_couplers)) != 1:
            raise ValueError(f"Components {component_1} and {component_2} have {n_couplers} connecting couplers.")
        return next(iter(connecting_couplers))

    def get_neighbor_locus_components(self, components: Collection[str]) -> set[str]:
        """Chip components that are connected to the given components by a coupler, but not included in them.

        Args:
            components: some chip components, typically qubits and computational resonators
        Returns:
            components that are connected to ``components`` by a coupler, but not included in them

        """
        neighbor_components = set()
        for coupler in self.get_neighbor_couplers(components):
            neighbor_components |= set(self.coupler_to_components[coupler])
        return neighbor_components - set(components)

    def get_connected_probe_lines(self, components: Collection[str]) -> set[str]:
        """Get probelines that are connected to any of the given components."""
        return {self.component_to_probe_line[c] for c in components}

    @staticmethod
    def limit_values(dct: ComponentMap, limit_to: Collection[str]) -> ComponentMap:
        """Prunes the given dictionary (e.g. a coupler-to-qubits map) to a subset of values.

        Used to prune e.g. :attr:`coupler_to_components` to a subset of relevant elements.

        Args:
            dct: Dictionary of collections of values.
            limit_to: Components to limit the output to.

        Returns:
            The input dictionary, but only with key-value pairs where the value intersects with `limit_to`.

        """
        return {key: values for key, values in dct.items() if any(v in limit_to for v in values)}

    def is_qubit(self, component: str) -> bool:
        """True iff the given component is a qubit."""
        return component in self.qubits

    def is_coupler(self, component: str) -> bool:
        """True iff the given component is a coupler."""
        return component in self.couplers

    def is_probe_line(self, component: str) -> bool:
        """True iff the given component is a probe line."""
        return component in self.probe_lines

    def is_computational_resonator(self, component: str) -> bool:
        """True iff the given component is a computational resonator."""
        return component in self.computational_resonators

    def set_locus_mapping(self, name: str, mapping: dict[Locus, tuple[str, ...]]) -> None:
        """Add a custom mapping from a gate locus to a set of components required for the gate operation.

        The mapping is of the form {<locus>: <components mapped to locus>}, where a locus can be mapped to one or more
        components. The locus itself can be a frozenset (denoting a symmetric gate) or a tuple (non-symmetric gate).

        Some examples:
        - ``DEFAULT_2QB_MAPPING`` (added in :meth:`__init__`) maps pairs of qubits to their common coupler symmetrically.
        - Fast flux CZ-gate maps pairs of qubits to their couplers non-symmetrically (first locus qubit can perform flux pulses).
        - A two-qubit gate implementation that includes playing pulses on neighboring components in addition to the
          connecting coupler.

        Args:
            name: The name for the gate & implementation this locus mapping represents (typically in the format
                ``"<gate name>.<implementation name>"``).
            mapping: The locus mapping to be added.

        """  # noqa: E501
        self._validate_locus_mapping(mapping)
        self._locus_mappings[name] = mapping

    def _validate_locus_mapping(self, mapping: dict[str | tuple[str], Locus] | None = None) -> None:
        """Validate that the components given in mapping are found in self and the mapping is correctly formed."""
        for locus, mapped in mapping.items():
            if not isinstance(locus, tuple) and not isinstance(locus, frozenset):
                raise ValueError("Mapped loci need to be tuples or frozen sets of component names")
            for mapped_component in mapped:
                if mapped_component not in self.all_components:
                    raise ValueError(f"Mapped component {mapped_component} is not found in this ChipTopology.")
            for locus_component in locus:
                if locus_component not in self.all_components:
                    raise ValueError(f"Locus component {locus_component} is not found in this ChipTopology.")

    def map_locus(self, locus: Locus, name: str | None = None) -> str | tuple[str] | None:
        """Returns the mapped components for the given locus and the given gate.

        If the locus or the gate is not found from the locus mappings of self, returns None.

        Args:
            locus: The locus to map.
            name: The name for the gate & implementation with which to map the locus (typically in the format
                ``"<gate name>.<implementation name>"``).

        Returns:
            The components mapped to the given locus or `None` if locus is not found in the given mapping.

        """
        if not name:
            if len(locus) == 1:
                name = DEFAULT_1QB_MAPPING
            elif len(locus) == 2:
                name = DEFAULT_2QB_MAPPING
        if name not in self._locus_mappings:
            return None
        return self._locus_mappings[name].get(locus, None)

    def map_to_locus(self, mapped: str | tuple[str], name: str) -> Locus | None:
        """Returns the locus that is mapped to the given components.

        Args:
            mapped: The mapped components.
            name: The name for the gate & implementation with which to map the locus (typically in the format
                ``"<gate name>.<implementation name>"``).

        Returns:
            The locus mapped to the given components or `None` if the components are not mapped to any locus.

        """
        if name not in self._locus_mappings:
            return None
        for locus, mapped_locus in self._locus_mappings[name].items():
            if mapped_locus == mapped:
                return locus
        return None

    def get_loci(self, name: str, default_mapping_dimension: int | None = None) -> list[Locus]:
        """Gives all the loci of a given gate.

        If no mapping with the given the name nor a default mapping with the given dimensionality is found,
        returns an empty list.

        Args:
            name: The name for the gate & implementation with which to map the locus (typically in the format
                ``"<gate name>.<implementation name>"``).
            default_mapping_dimension: If provided, will return the loci of the default mapping of the given
                dimensionality in case no mapping for ``name`` can be found.

        Returns:
            The loci associated with the given gate.

        """
        if name not in self._locus_mappings:
            if default_mapping_dimension == 1:
                name = DEFAULT_1QB_MAPPING
            elif default_mapping_dimension == 2:
                name = DEFAULT_2QB_MAPPING
        return [locus for locus in self._locus_mappings.get(name, {})]

    def get_common_computational_resonator(self, first_qubit: str, second_qubit: str) -> str:
        """Convenience method for getting the name of a computational resonator which is connected to both specified
        qubit components via tunable couplers.

        Args:
             first_qubit: The name of the first qubit.
             second_qubit: The name of the second qubit.
         The order of qubits does not matter, i.e. the `first_qubit` and `second_qubit` arguments are interchangeable.

        Returns:
             - The name of the computational resonator that is connected to both inputted qubits via tunable couplers.

        Raises:
             - ValueError: If no computational resonator was found that is connected to both qubits via tunable
             couplers.

        """
        neighbor_components = list(self.get_neighbor_locus_components([first_qubit, second_qubit]))  # noqa: F841

        resonators = [
            r
            for r in self.get_neighbor_locus_components([first_qubit, second_qubit])
            if r in self.computational_resonators
        ]
        common_resonators = [
            r
            for r in resonators
            if len(self.get_connecting_couplers([first_qubit, r])) == 1
            and len(self.get_connecting_couplers([second_qubit, r])) == 1
        ]

        if len(common_resonators) == 0:  # if no computational resonator is connected to both qubits
            raise ValueError(
                f"No computational resonator was found, that is connected to both qubits {first_qubit} and "
                f"{second_qubit} via tunable couplers."
            )
        if (
            len(common_resonators) == 1
        ):  # if only one computational resonator is connected to both qubits via tunable couplers
            computational_resonator = common_resonators[0]
        else:
            computational_resonator = sorted(common_resonators)[0]
            print(
                f"Warning: There was no unique computational resonator found, which connects to both qubits"
                f"{first_qubit} and {second_qubit} via tunable couplers. Use first one found: "
                f"{computational_resonator}."
            )
        return computational_resonator

    def get_all_common_resonators(
        self,
        qubits: list[str],
    ) -> set[str]:
        """Computational resonators connected to all the given qubits via a coupler.

        Args:
            qubits: Qubit names.

        Returns:
            Names of the computational resonators neighboring all of ``qubits`` (can be an empty set).

        """
        if not qubits:
            return set()
        common_resonator_set = set.intersection(*(self.get_neighbor_locus_components([qubit]) for qubit in qubits))
        # ensure the resonator set contains only resonators
        return common_resonator_set.intersection(self.computational_resonators)



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import logging.config
from typing import Any, Callable, Optional

BRIEF_DATEFMT = "%m-%d %H:%M:%S"

BRIEF = "[{asctime};{levelname:.1}{extra_info}] {message}"
VERBOSE = "[{asctime};{levelname};{processName}({process});{threadName}({thread});{name};{filename}:{lineno}:{funcName}{extra_info}] {message}"  # noqa: E501


class ExtraFormatter(logging.Formatter):
    """Helper formatter class to pass in arbitrary extra information to log messages."""

    def __init__(self, *args, extra_info_getter: Optional[Callable[[], str]] = None, **kwargs):
        self.extra_info_getter = extra_info_getter if extra_info_getter is not None else lambda: ""
        super().__init__(*args, **kwargs)

    def format(self, record):
        extra_info = self.extra_info_getter()
        record.__dict__.update(extra_info=extra_info)
        return super().format(record)


class InfoFilter(logging.Filter):
    """Helper class to filter log messages above INFO level."""

    def filter(self, record: logging.LogRecord) -> bool:
        return record.levelno in (logging.DEBUG, logging.INFO)


def init_loggers(
    loggers: Optional[dict[str, Optional[str]]] = None,
    default_level: str = "INFO",
    verbose: bool = False,
    extra_info_getter: Optional[Callable[[], str]] = None,
) -> None:
    """Set the log level of given logger names.

    Logs with INFO or DEBUG logging levels are written into stdout, and logs with other levels are written into stderr.

    By default, the root logger uses WARNING level.

    See Python's logging module for possible logging levels.

    Args:
        loggers: A mapping from logger name to (case insensitive) logging level. If logging level is None or empty
            string, ``default_level`` will be used for that logger. It is possible to fine tune logging for individual
            modules, since logger propagation is turned off. Overriding works both ways: a parent logger can have higher
            logging level than its descendants, and vice versa.
            For example, ``{"sqlalchemy": None, "sqlalchemy.engine": "debug"}`` will put
            "sqlalchemy" and its descendants (i.e. "sqlalchemy.dialects", "sqlalchemy.dialects.sqlite", etc.) into
            ``default_level``, except for "sqlalchemy.engine" for which DEBUG level is defined.
            For the root logger, use empty string key, for example: ``{"": "INFO"}``.
            If empty dict or None is given, only the root logger will be initialized to its default WARNING level.
        default_level: The default level (case insensitive) to be used for logger names given with ``loggers``
            for which a level is not specified.
        verbose: If False, :const:``BRIEF`` format will be used for log messages, otherwise :const:``VERBOSE``.
        extra_info_getter: Optional callable to convey extra information to log messages. It will get called before
            each log message emission and the output will get appended to the log message.

    """
    loggers = loggers or {}

    lowest_level_num = logging.CRITICAL
    loggers_config: dict[str, Any] = {}
    for logger_name, level in loggers.items():
        if not level:
            level = default_level  # noqa: PLW2901
        level = level.upper()  # noqa: PLW2901
        level_num = logging.getLevelName(level)
        if not isinstance(level_num, int):
            raise ValueError(f"Logger name '{logger_name}' has an incorrect level value '{level}'.")
        lowest_level_num = min(level_num, lowest_level_num)
        loggers_config[logger_name] = {
            "level": level,
            "handlers": ["stdout", "stderr"],
            "propagate": False,
        }
    loggers_config.setdefault("", {"level": "WARNING", "handlers": ["stderr"]})

    lowest_level = logging.getLevelName(lowest_level_num)
    lowest_level_warning_or_above = lowest_level if lowest_level_num > logging.WARNING else "WARNING"

    config = {
        "version": 1,
        "filters": {
            "infofilter": {
                "()": InfoFilter,
            }
        },
        "disable_existing_loggers": True,
        "formatters": {
            "brief": {
                "()": lambda: ExtraFormatter(
                    fmt=BRIEF, datefmt=BRIEF_DATEFMT, style="{", extra_info_getter=extra_info_getter
                )
            },
            "verbose": {"()": lambda: ExtraFormatter(fmt=VERBOSE, style="{", extra_info_getter=extra_info_getter)},
        },
        "handlers": {
            "stdout": {
                "level": lowest_level,
                "class": "logging.StreamHandler",
                "formatter": "verbose" if verbose else "brief",
                "stream": "ext://sys.stdout",
                "filters": ["infofilter"],
            },
            "stderr": {
                "level": lowest_level_warning_or_above,
                "class": "logging.StreamHandler",
                "formatter": "verbose" if verbose else "brief",
                "stream": "ext://sys.stderr",
            },
        },
        "loggers": loggers_config,
    }
    logging.config.dictConfig(config)



from typing import Any, Self

import pydantic
from pydantic import ConfigDict


class BaseModel(pydantic.BaseModel):
    """Pydantic base model to change the behaviour of pydantic globally.
    Note that setting model_config in child classes will merge the configs rather than override this one.
    https://docs.pydantic.dev/latest/concepts/config/#change-behaviour-globally
    """

    model_config = ConfigDict(
        # extra="forbid",  # Forbid any extra attributes
        # TODO (Marko): Use "ignore" at least temporarily
        #  Data sent by old server code might include extra attributes like "parent_name",
        #  which fails with a new client using extra="ignore".
        #  Before merging to master, consider if we want to be strict with extra="forbid",
        #  or would it be better to use extra="ignore" (shouldn't be needed if we update the server code anyway).
        extra="ignore",  # Ignore any extra attributes
        validate_assignment=True,  # Validate the data when the model is changed
        validate_default=True,  # Validate default values during validation
        ser_json_inf_nan="constants",  # Will serialize infinity and NaN values as Infinity and NaN
        frozen=True,  # This makes instances of the model potentially hashable if all the attributes are hashable
    )

    def model_copy(self, *, update: dict[str, Any] | None = None, deep: bool = True) -> Self:
        """Returns a copy of the model.

        Overrides the Pydantic default 'model_copy' to set 'deep=True' by default.
        """
        return super().model_copy(update=update, deep=deep)

    def copy(self, **kwargs) -> Self:
        """Returns a copy of the model.

        DEPRECATED: Use model_copy(update: dict[str, Any], deep: bool) instead.
        """
        # Call deprecated copy() here deliberately to trigger deprecation warning from Pydantic.
        return super().copy(update=kwargs, deep=True)



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convert numpy arrays to protos and back."""

import iqm.data_definitions.common.v1.data_types_pb2 as dpb
import numpy as np


def pack(array: np.ndarray) -> dpb.Array:
    """Packs a numeric numpy array into protobuf format.

    Args:
        array: Numpy array to convert.

    Returns:
        A protobuf instance that encapsulates `array`.

    """
    target = dpb.Array()
    target.shape.MergeFrom(array.shape)
    if not array.size:  # MergeFrom throws with 0-sized iterables
        return target

    dtype_type = array.dtype.type
    if dtype_type == np.complex128:
        target.complex128_array.real.MergeFrom(map(np.ndarray.item, np.nditer(array.real, order="C")))  # type:ignore
        target.complex128_array.imag.MergeFrom(map(np.ndarray.item, np.nditer(array.imag, order="C")))  # type:ignore
        return target

    if dtype_type == np.bool_:
        target_field = target.bool_array
    elif dtype_type == np.int64:
        target_field = target.int64_array
    elif dtype_type == np.int32:
        target_field = target.int64_array
    elif dtype_type == np.float64:
        target_field = target.float64_array
    else:
        raise TypeError(f"Unsupported numpy array type {dtype_type} for an array.")

    target_field.items.MergeFrom(map(np.ndarray.item, np.nditer(array, order="C")))  # type:ignore
    return target


def unpack(source: dpb.Array) -> np.ndarray:
    """Unpacks protobuf to array. Reverse operation of :func:`.pack`.

    Args:
        source: A protobuf instance that encapsulates some data.

    Returns:
        Unpacked data.

    Raises:
        ValueError or google.protobuf.message.DecodeError in case of invalid buffer

    """
    kind = source.WhichOneof("kind")
    if kind is None:
        return np.array([])

    shape = tuple(source.shape)
    if kind == "complex128_array":
        size = np.prod(shape).item()
        array = np.empty((size,), dtype=complex)
        array.real = np.fromiter(source.complex128_array.real, dtype=np.float64, count=size)
        array.imag = np.fromiter(source.complex128_array.imag, dtype=np.float64, count=size)
    elif kind == "bool_array":
        array = np.fromiter(source.bool_array.items, np.bool_)
    elif kind == "int64_array":
        array = np.fromiter(source.int64_array.items, np.int64)
    elif kind == "float64_array":
        array = np.fromiter(source.float64_array.items, np.float64)
    else:
        raise TypeError(f"Cannot unpack value in field {kind}. Field name not recognized.")

    return array.reshape(*shape)



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper to create a SweepOptions instance from a dict."""

import ast
from typing import Any

from exa.common.control.sweep.option import (
    CenterSpanBaseOptions,
    CenterSpanOptions,
    FixedOptions,
    StartStopBaseOptions,
    StartStopOptions,
)
from exa.common.control.sweep.option.constants import OPTIONS_TYPE
from exa.common.control.sweep.option.sweep_options import SweepOptions


def convert_to_options(config: dict[str, Any]) -> SweepOptions:
    """Creates one of the options object based on configuration dictionary.

    - If configuration has keys ``start`` and ``stop``, :class:`.StartStopOptions` is created.
    - If configuration has keys ``start_exp`` and ``stop_exp``, :class:`.StartStopBaseOptions` is created.
    - If configuration has keys ``center`` and ``span``, :class:`.CenterSpanOptions` is created.
    - If configuration has keys ``center_exp`` and ``span_exp``, :class:`.CenterSpanBaseOptions` is created
    - If configuration has keys ``fixed``, :class:`.FixedOptions` is created.

    Args:
        config: Configuration dictionary.

    Raises:
        ValueError: Error is raised if config has unsupported structure

    """
    config = {k.lower(): v for k, v in config.items()}
    if {OPTIONS_TYPE.get("start"), OPTIONS_TYPE.get("stop")}.issubset(set(config)):
        updated_config = __update_config(config, OPTIONS_TYPE.get("start_stop_list"))
        return StartStopOptions(**updated_config)
    elif {OPTIONS_TYPE.get("start_exp"), OPTIONS_TYPE.get("stop_exp")}.issubset(set(config)):
        __rename_key(config, OPTIONS_TYPE.get("start_exp"), OPTIONS_TYPE.get("start"))
        __rename_key(config, OPTIONS_TYPE.get("stop_exp"), OPTIONS_TYPE.get("stop"))
        updated_config = __update_config(config, OPTIONS_TYPE.get("start_stop_base_list"))
        return StartStopBaseOptions(**updated_config)
    elif {OPTIONS_TYPE.get("center"), OPTIONS_TYPE.get("span")}.issubset(set(config)):
        updated_config = __update_config(config, OPTIONS_TYPE.get("center_span_list"))
        return CenterSpanOptions(**updated_config)
    elif {OPTIONS_TYPE.get("center_exp"), OPTIONS_TYPE.get("span_exp")}.issubset(set(config)):
        __rename_key(config, OPTIONS_TYPE.get("center_exp"), OPTIONS_TYPE.get("center"))
        __rename_key(config, OPTIONS_TYPE.get("span_exp"), OPTIONS_TYPE.get("span"))
        updated_config = __update_config(config, OPTIONS_TYPE.get("center_span_base_list"))
        return CenterSpanBaseOptions(**updated_config)
    elif OPTIONS_TYPE.get("fixed") in set(config):
        updated_config = __update_config(config, OPTIONS_TYPE.get("fixed"))
        return FixedOptions(**updated_config)
    else:
        raise ValueError(f"Config {config} cannot be converted to range options")


def __update_config(config: dict[str, Any], keys: list[str]) -> dict[str, Any]:
    return {key: ast.literal_eval("".join(str(value).strip())) for key, value in config.items() if key in keys}


def __rename_key(config: dict[str, Any], old_key: str, new_key: str) -> None:
    config[new_key] = config.pop(old_key)



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
from http import HTTPStatus
import io
import locale
from pathlib import Path
from urllib.parse import urlparse
from urllib.request import url2pathname

from requests import Response
from requests.adapters import BaseAdapter
from six import BytesIO


class FileAdapter(BaseAdapter):
    def __init__(self, set_content_length=True):
        super(FileAdapter, self).__init__()
        self._set_content_length = set_content_length

    def send(self, request, **kwargs):
        """Sends PreparedRequest object. Returns Response object."""
        # Check that the method makes sense. Only support GET
        if request.method not in ("GET", "HEAD"):
            raise ValueError("Invalid request method %s" % request.method)

        url_parts = urlparse(request.url)

        path = url_parts.netloc + f"{str(Path(url2pathname(url_parts.path)))}.json"
        response = Response()
        try:
            response.status_code = HTTPStatus.OK
            response.raw = io.open(path, "rb")
            response.url = request.url
        except IOError as err:
            if err.errno == errno.ENOENT:
                response.status_code = HTTPStatus.NOT_FOUND
            else:
                response.status_code = HTTPStatus.BAD_REQUEST

            # Wrap the error message in a file-like object
            # The error message will be localized, try to convert the string
            # representation of the exception into a byte stream
            resp_str = str(err).encode(locale.getpreferredencoding(False))
            response.raw = BytesIO(resp_str)

        response.raw.release_conn = response.raw.close

        return response

    def close(self):
        pass



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pydantic models for CHAD."""

from collections.abc import Collection
from functools import cached_property
import re
from typing import Union

from pydantic import Field, field_validator

from exa.common.qcm_data.immutable_base_model import ImmutableBaseModel


def _natural_sort_key(name: str) -> tuple[int, ...]:
    return tuple(int(item) if item.isdigit() else item.lower() for item in re.split(r"(\d+)", name))


class Component(ImmutableBaseModel):
    name: str
    connections: tuple[str, ...] = ()

    def __lt__(self, other):
        return _natural_sort_key(self.name) < _natural_sort_key(other.name)

    @field_validator("connections", mode="before")
    @classmethod
    def sort(cls, connections: tuple[str, ...]) -> tuple[str, ...]:
        return tuple(sorted(connections, key=_natural_sort_key))


class Qubit(Component):
    pass


class Coupler(Component):
    pass


class ProbeLine(Component):
    pass


class Launcher(Component):
    pin: str
    function: str


class ComputationalResonator(Component):
    pass


class Components(ImmutableBaseModel):
    # Alias list of components to plurals rather than singulars to follow normal naming conventions
    qubits: tuple[Qubit, ...] = Field((), alias="qubit")
    couplers: tuple[Coupler, ...] = Field((), alias="tunable_coupler")
    probe_lines: tuple[ProbeLine, ...] = Field((), alias="probe_line")
    launchers: tuple[Launcher, ...] = Field((), alias="launcher")
    computational_resonators: tuple[ComputationalResonator, ...] = Field((), alias="computational_resonator")

    @field_validator("*")
    @classmethod
    def sort_components(cls, components: tuple[Component, ...]) -> tuple[Component, ...]:
        return tuple(sorted(components))

    @cached_property
    def all(self) -> dict[str, Component]:
        components: tuple[Qubit, Coupler, ProbeLine, Launcher, ComputationalResonator] = (
            self.qubits + self.couplers + self.probe_lines + self.launchers + self.computational_resonators
        )
        return {component.name: component for component in components}


class CHAD(ImmutableBaseModel):
    mask_set_name: str
    variant: str
    components: Components

    def __init__(self, **kwargs):
        kwargs["components"] = kwargs.pop("content")["components"]
        super().__init__(**kwargs)

    def get_component(self, component_name: str) -> Component:
        """Get component by component name."""
        return self.components.all[component_name]

    @cached_property
    def qubit_names(self) -> list[str]:
        """Names of all the qubits declared in CHAD data."""
        return self._get_component_names(self.components.qubits)

    @cached_property
    def coupler_names(self) -> list[str]:
        """Names of all the couplers declared in CHAD data."""
        return self._get_component_names(self.components.couplers)

    @cached_property
    def probe_line_names(self) -> list[str]:
        """Names of all the probe lines declared in CHAD data."""
        return self._get_component_names(self.components.probe_lines)

    @cached_property
    def computational_resonator_names(self) -> list[str]:
        """Names of all the computational resonators declared in CHAD data."""
        return self._get_component_names(self.components.computational_resonators)

    # TODO: Consider more generic "filter_components(components, component_type)" approach
    def filter_qubit_components(self, component_names: Collection[str]) -> list[str]:
        """Filter qubit components from the input components."""
        self._validate_input(component_names)
        return [component_name for component_name in component_names if component_name in self.qubit_names]

    def get_probe_line_names_for(self, component_names: Collection[str]) -> list[str]:
        """Get probe lines for given qubits in CHAD data."""
        self._validate_input(component_names)
        qubits = (qubit for qubit in self.components.qubits if qubit.name in component_names)

        probe_line_names: set[str] = set()
        for qubit in qubits:
            for probe_line in self.components.probe_lines:
                if probe_line.name in qubit.connections:
                    probe_line_names.add(probe_line.name)

        return list(probe_line_names)

    def group_components_per_default_operations(
        self,
        component_names: Collection[str],
    ) -> tuple[dict[str, list[str]], dict[str, list[str]]]:
        """Groups given qubits/couplers based on their defined default operations.

        The default operations that can be deducted from a CHAD are `readout`, `drive`, and `flux`.

        Args:
            component_names: The component names to which to do the grouping.
              Other components present in the CHAD will not be included in the returned data.

        Returns:
            - Tuple of qubits and couplers mapped to their connected default operations.
                The data is in the form of a dict with the keys being `readout`, `drive`, and `flux`,
                and the values the list of component names having that particular operation.

        """
        self._validate_input(component_names)

        def format_output(_types_and_operations: dict) -> tuple[dict, dict]:
            _qubits = {key: list(value) for key, value in _types_and_operations["qubit"].items()}
            _couplers = {key: list(value) for key, value in _types_and_operations["coupler"].items()}
            return _qubits, _couplers

        types_and_operations: dict[str, dict[str, set[str]]] = {
            "qubit": {"readout": set(), "drive": set(), "flux": set()},
            "coupler": {"readout": set(), "drive": set(), "flux": set()},
        }

        launchers = {launcher.name: launcher for launcher in self.components.launchers}

        component: Union[Qubit, Coupler]
        for component in [*self.components.qubits, *self.components.couplers]:
            component_type = "qubit" if isinstance(component, Qubit) else "coupler"
            if component.name in component_names:
                for connection in component.connections:
                    if connection in self.probe_line_names:
                        types_and_operations[component_type]["readout"].add(component.name)
                    elif connection in launchers and launchers[connection].function == "drive":
                        types_and_operations[component_type]["drive"].add(component.name)
                    elif connection in launchers and launchers[connection].function == "flux":
                        types_and_operations[component_type]["flux"].add(component.name)
        return format_output(types_and_operations)

    def get_coupler_mapping_for(self, component_names: Collection[str]) -> dict[str, list[str]]:
        """Get the coupler-component mapping for the couplers that connects to at least two components
             in the given qubits.

        Args:
            component_names: The qubit names. May contain any number of qubits.

        Returns:
            Coupler names mapped to the components they connect.

        Raises:
            - ValueError: If the provided qubit name list contains duplicates.

        """
        self._validate_input(component_names)

        component_names = list(component_names) + self.computational_resonator_names
        component_mapping: dict[str, list[str]] = {}
        for coupler in self.components.couplers:
            connections = [
                component_name for component_name in component_names if component_name in coupler.connections
            ]
            if len(connections) > 1:
                component_mapping[coupler.name] = connections
        return component_mapping

    def get_probe_line_mapping_for(self, component_names: Collection[str]) -> dict[str, list[str]]:
        """Get the probe line-component mapping.

        Args:
            component_names: The qubit names. May contain any number of qubits.

        """
        self._validate_input(component_names)

        component_mapping: dict[str, list[str]] = {}
        for probe_line in self.components.probe_lines:
            connections = [
                component_name for component_name in component_names if component_name in probe_line.connections
            ]
            if connections:
                component_mapping[probe_line.name] = connections
        return component_mapping

    def get_common_coupler_for(self, first_component: str, second_component: str) -> str:
        """Convenience method for getting the name of a coupler connecting a pair of components.

        Args:
            first_component: The name of the first component.
            second_component: The name of the second component.
                The order of qubits does not matter, i.e. the `first_qubit` and `second_qubit`
                arguments are interchangeable.

        Returns:
            - The name of the coupler that connects the inputted components.

        Raises:
            - ValueError: If there were no couplers or more than one coupler connecting the component pair (the latter
                should not be possible in a realistic chip).

        """
        coupler_mapping = self.get_coupler_mapping_for([first_component, second_component])
        common_couplers = [
            coupler
            for coupler, components in coupler_mapping.items()
            if all(component in components for component in [first_component, second_component])
        ]
        if len(common_couplers) != 1:
            raise ValueError(
                f"No common coupler was found for {first_component} and {second_component} "
                f"or there were multiple couplers. "
                f"Check your Chip Architecture Definition or the inputted component names.\n"
                f"Found common couplers: {common_couplers}"
            )
        return common_couplers[0]

    @staticmethod
    def _get_component_names(components: tuple[Component, ...]) -> list[str]:
        return [component.name for component in components]

    def _validate_input(self, component_names: Collection[str]) -> None:
        if len(set(component_names)) < len(component_names):
            raise ValueError(f"The provided component names {component_names} contain duplicates.")
        chad_component_names = self.components.all.keys()
        for component_name in component_names:
            if component_name not in chad_component_names:
                raise ValueError(f"The provided component name '{component_name}' doesn't exist in the CHAD.")
        # TODO: Validate only certain type of components are given



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QCM (Quantum Computer Management) Data API client implementation."""

from collections.abc import Callable
from functools import cache
import json
import logging
from pathlib import Path
from typing import Any

from packaging.version import Version
import requests

from exa.common.errors.exa_error import ExaError, RequestError
from exa.common.qcm_data.file_adapter import FileAdapter

MIN_SUPPORTED_CONTENT_FORMAT_VERSION = Version("1.0")
MAX_SUPPORTED_CONTENT_FORMAT_VERSION = Version("2")  # Only major is checked for max version. Minor has no effect.


class QCMDataClient:
    """Python client for QCM (Quantum Computer Management) Data API.

    Args:
        root_url: URL pointing to QCM Data service.
            This URL can point to a local file storage as well.
            In that case, the URL should point to a directory which
            has a directory structure identical to QCM Data service (for example /chip-data-records/),
            and files containing data in identical format returned by QCM Data service.
            For example, CHAD files should be named {chip_label}.json, like M156_W531_A09_L09.json, and contain
            a list instead of a single object.
        fallback_root_url: Same as `root_url`, used if a query via `root_url` returns nothing.

    """

    def __init__(self, root_url: str, fallback_root_url: str = ""):
        self.logger = logging.getLogger(__name__ + "." + self.__class__.__name__)
        self.logger.info("Initialize QCMDataClient: root_url=%s", root_url)

        if not root_url:
            raise ValueError("QCMDataClient 'root_url' cannot be empty, it must be a valid HTTP or file URL.")
        self._root_url = root_url
        self._fallback_root_url = fallback_root_url

        self.session = requests.Session()
        self.session.mount(str(Path("file://")), FileAdapter())

        # Make the cache containers local to the instances so that the reference from cache to the instance
        # gets scraped off with the instance
        # https://rednafi.github.io/reflections/dont-wrap-instance-methods-with-functoolslru_cache-decorator-in-python.html
        self._send_request = cache(self._send_request)

    @property
    def root_url(self) -> str:
        """Returns the remote QCM Data service URL."""
        return self._root_url

    @root_url.setter
    def root_url(self, root_url: str) -> None:
        """Sets the remote QCM Data service URL.

        Cache for :func:`get_chad` has to be reset when the URL changes.
        """
        if self._root_url != root_url:
            self._root_url = root_url

    def get_chip_design_record(self, chip_label: str) -> dict:
        """Get a raw chip design record matching the given chip label.

        Args:
            chip_label: Chip label.

        Returns:
            Data record matching the given chip label.

        """
        url_tail = f"/cheddars/{chip_label}?target=in-house"
        try:
            response = self._send_request(self.session.get, f"{self.root_url}{url_tail}")
        except RequestError as err:
            if self._fallback_root_url:
                response = self._send_request(self.session.get, f"{self._fallback_root_url}{url_tail}")
            else:
                raise err
        data = response.json().get("data")
        if not data:
            raise RequestError(f"Chip design record for {chip_label} does not contain data.")
        self._validate_chip_design_record(data, chip_label)
        return data

    def _send_request(self, http_method: Callable[..., requests.Response], url: str) -> requests.Response:
        # Send the request and return the response.
        # Will raise an error if respectively an error response code is returned.
        # http_method should be any of session.[post|get|put|head|delete|patch|options]
        self.logger.debug("http_method=%s, url=%s", http_method.__name__, url)
        response = http_method(url, timeout=20)
        if not response.ok:
            try:
                response_dict = response.json()
                error_message = response_dict["detail"]
            except json.JSONDecodeError:
                error_message = response.text
            raise RequestError(f"{url} returned error code {response.status_code}: {error_message}")
        return response

    @staticmethod
    def _validate_chip_design_record(chip_design_record: dict[str, Any], chip_label: str) -> None:
        chip_label_mask_set_name, _, chip_label_variant, _ = chip_label.split("_")

        content_format_version = Version(chip_design_record["content_format_version"])
        chad_mask_set_name = chip_design_record["mask_set_name"]
        chad_variant = chip_design_record["variant"]

        # Allow any release in version range.
        if (
            content_format_version < MIN_SUPPORTED_CONTENT_FORMAT_VERSION
            or content_format_version.major > MAX_SUPPORTED_CONTENT_FORMAT_VERSION.major
        ):
            raise ExaError(f"CHAD content format version '{content_format_version.public}' is not supported.")

        if chad_mask_set_name != chip_label_mask_set_name:
            raise ExaError(
                f"CHAD mask set name '{chad_mask_set_name}' doesn't match the chip label mask set name "
                f"'{chip_label_mask_set_name}'."
            )

        if chad_variant != chip_label_variant:
            raise ExaError(
                f"CHAD variant '{chad_variant}' doesn't match the chip label variant '{chip_label_variant}'."
            )



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Range specification for arbitrary set of values."""

from dataclasses import dataclass

import numpy as np

from exa.common.control.sweep.option.sweep_options import SweepOptions
from exa.common.control.sweep.sweep_values import SweepValues


@dataclass(frozen=True)
class FixedOptions(SweepOptions):
    """Range fixed options."""

    #: List of values.
    fixed: SweepValues

    @property
    def data(self) -> SweepValues:
        return np.asarray(self.fixed).tolist()



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A tree-structured container for :class:`Settings <exa.common.data.parameter.Setting>`.

The :class:`.SettingNode` class combines a bunch of Settings together.
It may also contain other SettingNodes.
Together, the contents form a tree structure that provides a useful way of grouping Settings.

As an example, we manually construct a tree of SettingNodes with some dummy Settings, but it is usually not necessary.
The root node in the following examples is called ``'node'``.

.. testsetup:: pulse

    from exa.common.data.parameter import Setting, Parameter
    from exa.common.data.setting_node import SettingNode
    node = SettingNode('root',
        flux=SettingNode('root.flux',
            voltage=Setting(Parameter('root.flux.voltage', 'Voltage', 'V'), 1.5),
            resistance=Setting(Parameter('root.flux.resistance', 'Resistance', 'Ohm'), None),
        ),
        pulse=SettingNode('root.pulse',
            amplitude=Setting(Parameter('root.pulse.amplitude', 'Amplitude'), 1.0),
            duration=Setting(Parameter('root.pulse.duration', 'Duration', 's'), 100e-9),
       )
    )


What's inside?
--------------

The easiest way to see the content of the node is the :meth:`.SettingNode.print_tree` method:

.. doctest:: pulse
   :options: +NORMALIZE_WHITESPACE

    >>> node.print_tree(levels=1)
     "root"
     ║
     ╚═ flux: "root.flux"
     ╚═ pulse: "root.pulse"


We see that the ``'root'`` node has two children, named ``'root.flux'`` and ``'root.pulse'``, which
themselves are also SettingNodes.
This follows the typical naming convention in EXA: Subnodes include the names of their parents, separated by a dot.

.. doctest:: pulse
   :options: +NORMALIZE_WHITESPACE

    >>> node.print_tree()
     "root"
     ║
     ╠═ flux: "root.flux"
     ║   ╠─ voltage: Voltage = 1.5 V
     ║   ╚─ resistance: Resistance = None (automatic/unspecified)
     ╚═ pulse: "root.pulse"
         ╠─ amplitude: Amplitude = 1.0
         ╚─ duration: Duration = 1e-07 s


The children contain some dummy Settings, showing the keys, labels and current values.

For other ways to access the content of the node, see also :attr:`.SettingNode.children`,
:attr:`.SettingNode.all_settings`, and :meth:`.SettingNode.nodes_by_type`.


Get and set values
------------------

The values within the nodes can be accessed using the attribute or dictionary syntax:


.. doctest:: pulse

    >>> node.pulse.amplitude.value
    1.0
    >>> node['flux']['voltage'].value
    1.5

The values can be changed with a simple ``=`` syntax:

.. doctest:: pulse

    >>> node.pulse.amplitude = 1.4
    >>> node.pulse.amplitude.value
    1.4

.. note::

    ``node.setting`` refers to the Setting object. ``node.setting.value`` syntax refers to the data stored inside.

``SettingNode`` also supports "the path notation" by default (but not if ``align_name`` is set to ``False``,
since it cannot be made to work consistently if nodes are allowed to be named differently from their paths):

.. doctest:: pulse

    >>> node['flux.voltage']

is the same as ``node['flux']['voltage']``.

Basic manipulation
------------------

Adding and deleting new Settings and nodes is simple:

.. doctest:: pulse

    >>> modified = node.copy()
    >>> del modified.flux # removes the node
    >>> del modified.pulse.amplitude # removes the Setting
    >>> modified.pulse.my_new_setting = Setting(Parameter('my name'), 33)

It is usually a good idea to make a copy of the original node, so that it won't be modified accidentally.

The path notation of ``SettingNode``also works when inserting:

.. doctest:: pulse

    >>> node['flux.my.new.path.foo'] = Setting(Parameter('foo'), 1.0)

Any nodes that did not already exist under ``node`` will be inserted (in this case ``flux`` already existed, but
the rest not, so under ``flux`` the nodes ``my``, ``new``, and ``path`` would be added), and then finally the
value is added as child to the final node. Note: ``SettingNode`` always alings the path and name of any nodes under it,
so this would result in the new setting being renamed as "flux.my.new.path.foo":

.. doctest:: pulse

    >>> node['flux.my.new.path.foo'] = Setting(Parameter('bar'), 1.0)

If ``align_name`` is set to ``False", the name and path of nodes are not automatically aligned, but otherwise the above
path notation will still work. The added nodes will be named by just their path fragments ("my", "new", "path", and
so on), and the Setting will be added under the key "foo", but it will still retain its name "bar". Note: the root node
name will always be excluded from the paths (and names when they are aligned with the path), so that the path of
``root.foo.bar`` is ``"foo.bar"``.

To merge values of two SettingNodes, there are helpers :meth:`.SettingNode.merge` and
:meth:`.SettingNode.merge_values`.

The first one merges the tree structure and values of two nodes and outputs a third one as a result.
``None`` values are always replaced by a proper value if such exists. In case of conflicting nodes or values,
the content of the first argument takes priority.

.. doctest:: pulse
    :options: +NORMALIZE_WHITESPACE

    >>> result = SettingNode.merge(node.flux, node.pulse)
    >>> result.print_tree()
     "root.flux"
     ╠─ amplitude: Amplitude = 1.4
     ╠─ duration: Duration = 1e-07 s
     ╚─ voltage: Voltage = 1.5 V


Note how the result has values from ``node.flux``, but also settings ``node.pulse`` that do not exist in ``node.flux``.

The :meth:`.SettingNode.merge_values` method is an in-place operation that only changes
the values of Settings that already exist in the node, if possible:

.. doctest:: pulse
    :options: +NORMALIZE_WHITESPACE

    >>> modified = node.copy()
    >>> modified.flux.voltage = 222
    >>> modified.flux.resistance = 333
    >>> node.merge_values(modified, prioritize_other=True)
    >>> node.print_tree()
     "root"
     ║
     ╠═ flux: "root.flux"
     ║   ╠─ voltage: Voltage = 222 V
     ║   ╚─ resistance: Resistance = 333 Ohm
     ╚═ pulse: "root.pulse"
         ╠─ amplitude: Amplitude = 1.4
         ╚─ duration: Duration = 1e-07 s

Sometimes, it is easier to collect values in a dictionary and set them all at once by using
:meth:`.SettingNode.set_from_dict`. The nested structure of the dictionary should match
the structure of the SettingNode. Keys that are not found in the tree are silently ignored, unless the ``strict``
flag is used.

.. doctest:: pulse
    :options: +NORMALIZE_WHITESPACE

    >>> values_to_set = {'flux': {'resistance': 0.001}, 'ignored_entry': 234}
    >>> node.set_from_dict(values_to_set)
    >>> node.flux.print_tree()
     "root.flux"
     ╠─ voltage: Voltage = 222 V
     ╚─ resistance: Resistance = 0.001 Ohm


"""

from __future__ import annotations

from collections.abc import Generator, ItemsView, Iterator
from copy import copy
from itertools import permutations
import logging
import numbers
import pathlib
from typing import Any, Iterable

import jinja2
import numpy as np

from exa.common.data.base_model import BaseModel
from exa.common.data.parameter import CollectionType, Parameter, Setting
from exa.common.errors.exa_error import UnknownSettingError
from exa.common.qcm_data.chip_topology import sort_components

logger = logging.getLogger(__name__)


def _fix_path_recursive(node: SettingNode | dict, path: str) -> SettingNode | dict:
    """Recursively travel the settings tree and fix the ``path``attribute (also aligns ``name``,
    based on the node type). Deep copies all the child nodes.
    """
    settings = {}
    subtrees = {}
    for key, setting in node.settings.items():
        child_path = f"{path}.{key}"
        update_dict = {"path": child_path}
        if node.align_name:
            update_dict["parameter"] = setting.parameter.model_copy(update={"name": child_path})
        settings[key] = setting.model_copy(update=update_dict)
    for key, subnode in node.subtrees.items():
        subtrees[key] = _fix_path_recursive(subnode, f"{path}.{key}")
    node_update_dict = {"path": path, "settings": settings, "subtrees": subtrees}
    if node.align_name:
        node_update_dict["name"] = path
    return node.model_copy(update=node_update_dict, deep=False)


class SettingNode(BaseModel):
    """A tree-structured :class:`.Setting` container.

    Each child of the node is a :class:`.Setting`, or another :class:`SettingNode`.
    Iterating over the node returns all children, recursively.
    Settings can be accessed by dictionary syntax or attribute syntax:

    .. doctest::

        >>> from exa.common.data.parameter import Parameter
        >>> from exa.common.data.setting_node import SettingNode
        >>> p1 = Parameter("voltage", "Voltage")
        >>> f1 = Parameter("frequency", "Frequency")
        >>> sub = SettingNode("sub", frequency=f1)
        >>> settings = SettingNode('name', voltage=p1)
        >>> settings.voltage.parameter is p1
        True
        >>> settings['voltage'].parameter is p1
        True
        >>> settings.voltage.value is None
        True
        >>> settings.voltage = 7  # updates to Setting(p1, 7)
        >>> settings.voltage.value
        7
        >>> settings["sub.frequency"] = 8
        >>> settings["sub.frequency"].value
        8

    Args:
        name: Name of the node.
        settings: Dict of setting path fraqment names (usually the same as the setting name) to the settings. Mostly
            used when deserialising and otherwise left empty.
        subtrees: Dict of child node path fraqment names (usually the same as the child node name) to the settings.
            Mostly used when deserialising and otherwise left empty.
        path: Optionally give a path for the node, by default empty.
        generate_paths: If set ``True``, all subnodes will get their paths autogenerated correctly. Only set to
            ``False`` if the subnodes already have correct paths set (e.g. when deserialising).
        kwargs: The children given as keyword arguments. Each argument must be a :class:`.Setting`,
            :class:`.Parameter`, or a :class:`SettingNode`. The keywords are used as the names of the nodes.
            Parameters will be cast into Settings with the value ``None``.

    """

    name: str
    settings: dict[str, Setting] = {}
    subtrees: dict[str, SettingNode] = {}
    path: str = ""

    align_name: bool = True

    def __init__(
        self,
        name: str,
        settings: dict[str, Any] | None = None,
        subtrees: dict[str, Any] | None = None,
        *,
        path: str = "",
        align_name: bool = True,
        generate_paths: bool = True,
        **kwargs,
    ):
        settings: dict[str, Any] = settings or {}
        subtrees: dict[str, Any] = subtrees or {}

        for key, child in kwargs.items():
            if isinstance(child, Setting):
                settings[key] = child
            elif isinstance(child, Parameter):
                settings[key] = Setting(parameter=child, value=None, path=key)
            elif isinstance(child, SettingNode):
                subtrees[key] = child
            else:
                raise ValueError(f"{key} should be a Parameter, Setting or a SettingNode, not {type(child)}.")
        super().__init__(
            name=name,
            settings=settings,
            subtrees=subtrees,
            path=path,
            align_name=align_name,
            **kwargs,
        )

        if generate_paths:
            self._generate_paths_and_names()

    def _generate_paths_and_names(self) -> None:
        """This method generates the paths and aligns the names when required."""
        for key, child in self.subtrees.items():
            update_path = f"{self.path}.{key}" if self.path else key
            self.subtrees[key] = _fix_path_recursive(child, update_path)
        for key, child in self.settings.items():
            update_path = f"{self.path}.{key}" if self.path else key
            if isinstance(child, Setting):
                update_dict = {"path": update_path}
                if self.align_name:
                    update_dict["parameter"] = child.parameter.copy(name=update_path)
                self[key] = child.model_copy(update=update_dict)
            elif self.align_name:
                self[key] = Setting(
                    parameter=child.model_copy(update={"name": update_path}), value=None, path=update_path
                )
        if self.path and self.align_name:
            self.name = self.path

    def __getattr__(self, key):
        if key == "settings":
            # Prevent infinite recursion. If settings actually exists, this method is not called anyway
            raise AttributeError
        if key in self.settings:
            return self.settings[key]
        if key in self.subtrees:
            return self.subtrees[key]
        raise UnknownSettingError(
            f'{self.__class__.__name__} "{self.name}" has no attribute {key}. Children: {self.children.keys()}.'
        )

    def __dir__(self):
        """List settings and subtree names, so they occur in IPython autocomplete after ``node.<TAB>``."""
        return [name for name in list(self.settings) + list(self.subtrees) if name.isidentifier()] + super().__dir__()

    def _ipython_key_completions_(self):
        """List items and subtree names, so they occur in IPython autocomplete after ``node[<TAB>``"""
        return [*self.settings, *self.subtrees]

    def __setattr__(self, key, value):
        """Overrides default attribute assignment to allow the following syntax: ``self.foo = 3`` which is
        equivalent to ``self.foo.value.update(3)`` (if ``foo`` is a :class:`.Setting`).
        """
        path = self._get_path(key)
        if isinstance(value, (Setting, Parameter)):
            if isinstance(value, Parameter):
                if self.align_name:
                    value = value.model_copy(update={"name": path})
                value = Setting(parameter=value, value=None, path=path)
            else:
                update_dict = {"path": path}
                if self.align_name:
                    update_dict["parameter"] = value.parameter.model_copy(update={"name": path})
                value = value.model_copy(update=update_dict)
            self.settings[key] = value
            self.subtrees.pop(key, None)
        elif isinstance(value, SettingNode):
            self.subtrees[key] = _fix_path_recursive(value, path)
            self.settings.pop(key, None)
        elif key != "settings" and key in self.settings:  # != prevents infinite recursion
            self.settings[key] = self.settings[key].update(value)
        else:
            self.__dict__[key] = value

    def __delattr__(self, key):
        if key in self.settings:
            del self.settings[key]
        elif key in self.subtrees:
            del self.subtrees[key]
        else:
            del self.__dict__[key]

    def __getitem__(self, item: str) -> Setting | SettingNode:
        """Allows dictionary syntax."""
        if len(item.split(".")) == 1:
            return self.__getattr__(item)
        return self.get_node_for_path(item)

    def __setitem__(self, key: str, value: Any) -> None:
        """Allows dictionary syntax."""
        path_fragments = key.split(".")
        node_key = path_fragments[0]
        remaining_path = ".".join(path_fragments[1:])

        if not remaining_path:
            self.__setattr__(key, value)
        elif node_key in self.children:
            try:
                self[node_key][remaining_path] = value
            except TypeError:
                raise ValueError(f"Path '{key}' is invalid: '{node_key}' is a setting, not a node.")
        else:
            if not isinstance(value, (SettingNode, Setting, Parameter)):
                raise ValueError(
                    f"Assigning value {value} to path {key} cannot be done when this path is not found in self."
                )
            self.add_for_path({path_fragments[-1]: value}, path=".".join(path_fragments[:-1]))

    def __delitem__(self, key):
        """Allows dictionary syntax."""
        self.__delattr__(key)

    def __iter__(self):
        """Allows breadth-first iteration through the tree."""
        yield self
        yield from iter(self.settings.values())
        for subtree in self.subtrees.values():
            yield from iter(subtree)

    def nodes_by_type(
        self,
        node_types: type | tuple[type, ...] | None = None,
        recursive: bool = False,
    ) -> Iterator:
        """Yields all nodes, filtered by given `node_types`.

        Used to find and iterate over nodes of specific types.

        Args:
            node_types: when iterating over the tree, yields only instances that match this type
                or any of the types in the tuple. By default, yields Settings and SettingNodes.
            recursive: If True, the search is carried recursively. If False, the search is limited to
                immediate child nodes.

        Returns:
             Iterator that yields the filtered nodes.

        """
        node_types = node_types or (Setting, SettingNode)
        iterable = self if recursive else self.children.values()
        return filter(lambda node: isinstance(node, node_types), iterable)

    def update_setting(self, setting: Setting) -> None:
        """Update an existing `Setting` in this tree.

        Args:
            setting: Setting that will replace an existing Setting with the same name. Or if the setting is an
                element-wise setting (i.e. it has a non-empty value of ``setting.element_indices``), the corresponding
                element will be updated in the collection.

        Raises:
            UnknownSettingError: If no setting is found in the children of this tree.

        """

        def list_assign(value, array, indices_list) -> None:
            sub_array = array
            for index in indices_list[:-1]:
                sub_array = sub_array[index]
            sub_array[indices_list[-1]] = value

        for branch in self.nodes_by_type(SettingNode, True):
            for key, item in branch.children.items():
                if setting.element_indices is None:
                    if item.name == setting.name:
                        branch[key] = setting.value
                        return
                elif isinstance(item, Setting) and item.name == setting.parent_name and item.element_indices is None:
                    parent_value = item.value.copy()
                    list_assign(setting.value, parent_value, setting.element_indices)
                    branch[key] = parent_value
                    return
        raise UnknownSettingError(
            f'No Setting with name {setting.name} was found in {self.__class__.__name__} "{self.name}".'
        )

    @property
    def all_settings(self) -> Generator[Setting]:
        """Yields all :class:`.Setting` instances inside this node, recursively."""
        yield from self.nodes_by_type(Setting, recursive=True)

    @property
    def children(self) -> dict[str, Setting | SettingNode]:
        """Dictionary of immediate child nodes of this node."""
        return {**self.settings, **self.subtrees}

    @property
    def child_settings(self) -> ItemsView[str, Setting]:
        """ItemsView of settings of this node."""
        return self.settings.items()

    @property
    def child_nodes(self) -> ItemsView[str, SettingNode]:
        """ItemsView of immediate child nodes of this node."""
        return self.subtrees.items()

    def get_parent_of(self, name: str) -> SettingNode:
        """Get the first SettingNode that has a Setting named `name`.

        Args:
            name: Name of the setting to look for.

        Returns:
            A SettingNode that has a child `name`.

        """
        for branch in self.nodes_by_type(SettingNode, recursive=True):
            for setting in branch.children.values():
                if setting.name == name:
                    return branch
        raise UnknownSettingError(f'{name} not found inside {self.__class__.__name__} "{self.name}".')

    def find_by_name(self, name: str) -> SettingNode | Setting | None:
        """Find first occurrence of Setting or SettingNode by name, by iterating recursively through all children.

        Args:
            name: Name of the Setting or SettingNode to look for.

        Returns:
            First found item, or None if nothing is found.

        """
        return next((item for item in self if item.name == name), None)

    @staticmethod
    def merge(
        first: SettingNode,
        second: SettingNode,
        merge_nones: bool = False,
        align_name: bool = True,
        deep_copy: bool = True,
    ) -> SettingNode:
        """Recursively combine the tree structures and values of two SettingNodes.

        In case of conflicting nodes,values in `first` take priority regardless of the replaced content in `second`.
        `None` values are not prioritized unless ``merge_nones`` is set to ``True``.

        Args:
            first: SettingNode to merge, whose values and structure take priority
            second: SettingNode to merge.
            merge_nones: Whether to merge also ``None`` values from ``first`` to ``second``.
            align_name: Whether to align the paths (and also names if ``second`` does not use ``align_name==False``)
                when merging the nodes. Should never be set ``False`` unless the paths in ``first`` already align with
                what they should be in ``second`` (setting it ``False`` in such cases can improve performance).
            deep_copy: Whether to deepcopy or just shallow copy all the sub-nodes. Set to ``False`` with high caution
                and understand the consequences.

        Returns:
            A new SettingNode constructed from arguments.

        """
        new = second.model_copy(deep=deep_copy)
        for key, item in first.settings.items():
            if merge_nones or item.value is not None:
                if align_name:
                    new[key] = item.copy(name=item.name.replace(f"{first.name}.", ""))
                else:
                    new.settings[key] = item.copy()
        for key, item in first.subtrees.items():
            item_copy = item.model_copy(update={"name": item.name.replace(f"{first.name}.", "")}, deep=deep_copy)
            subs = new if align_name else new.subtrees
            if key in new.subtrees:
                subs[key] = SettingNode.merge(item_copy, new[key])
            else:
                subs[key] = item_copy

        for key, item in first.__dict__.items():
            if key not in ["settings", "subtrees"]:
                new[key] = copy(item)
        return new

    def merge_values(self, other: SettingNode, prioritize_other: bool = False):
        """Recursively combine the values from another :class:`SettingNode` to this one.

        The resulting tree structure the same as that of self.

        Args:
            other: SettingNode to merge.
            prioritize_other: If True, will prioritize values in other. If False (default), only None values in self
                will be replaced.

        """
        for key, item in other.settings.items():
            if key in self.settings and (prioritize_other or (self[key].value is None)):
                self.settings[key] = Setting(self.settings[key].parameter, item.value)
        for key, item in other.subtrees.items():
            if key in self.subtrees:
                self.subtrees[key].merge_values(copy(item), prioritize_other)

    def prune(self, other: SettingNode) -> None:
        """Recursively delete all branches from this SettingNode that are not found in ``other``."""
        for key, node in self.subtrees.copy().items():
            if key not in other.subtrees:
                del self[key]
            else:
                self[key].prune(other[key])

    def print_tree(self, levels: int = 5) -> None:
        """Print a tree representation of the contents of this node.

        Args:
            levels: display this many levels, starting from the root.

        """

        def append_lines(node: SettingNode, lines: list[str], indents: list[bool]):
            indent = "".join([" ║  " if i else "    " for i in indents])
            if len(indents) < levels:
                for key, setting in node.settings.items():
                    if setting.value is None:
                        value = "None (automatic/unspecified)"
                    elif setting.parameter.collection_type == CollectionType.NDARRAY:
                        value = str(setting.value).replace("\n", "") + f" {setting.unit}"
                    else:
                        value = f"{setting.value} {setting.unit}"
                    lines.append(indent + f" ╠─ {key}: {setting.label} = {value}")
                if node.subtrees:
                    lines.append(indent + " ║ ")
                subtrees_written = 0
                for key, subtree in node.subtrees.items():
                    lines.append(indent + f' ╠═ {key}: "{subtree.name}: align_name={subtree.align_name}"')
                    if subtrees_written == len(node.subtrees) - 1:
                        lines[-1] = lines[-1].replace("╠", "╚")
                    append_lines(subtree, lines, indents + [subtrees_written < len(node.subtrees) - 1])
                    subtrees_written += 1
            lines[-1] = lines[-1].replace("╠", "╚")

        lines = [f'"{self.name}: align_name={self.align_name}"']
        append_lines(self, lines, [])
        print("\n", "\n".join(lines))

    def __eq__(self, other: SettingNode):
        return isinstance(other, SettingNode) and (
            (self.name, self.settings, self.subtrees) == (other.name, other.settings, other.subtrees)
        )

    def __repr__(self):
        return f"{self.__class__.__name__}{self.children}"

    def __str__(self):
        content = ", ".join(f"{key}={node.__class__.__name__}" for key, node in self.children.items())
        return f"{self.__class__.__name__}({content})"

    @classmethod
    def transform_node_types(cls, node: SettingNode) -> SettingNode:
        """Reduce any subclass of SettingNode and it's contents into instances of `cls`.

        Args:
            node: node to transform.

        Return:
            A new SettingNode with the same structure as the original, but where node instances are of type `cls`.

        """
        new = cls(name=node.name, **node.settings, align_name=node.align_name, generate_paths=False)
        for key, subnode in node.subtrees.items():
            new.subtrees[key] = cls.transform_node_types(subnode)
        return new

    def set_from_dict(self, dct: dict[str, Any], strict: bool = False) -> None:
        """Recursively set values to Settings, taking values from a dictionary that has similar tree structure.
        Keys that are not found in self are ignored, unless `strict` is True.

        Args:
            dct: Dictionary containing the new values to use.
            strict: If True, will raise error if `dct` contains a setting that is not found in `self`.

        Raises:
            UnknownSettingError: If the condition of `strict` happens.

        """
        for key, value in dct.items():
            if key not in self.children:
                error = UnknownSettingError(f"Tried to set {key} to {value}, but no such node exists in {self.name}.")
                logger.debug(error.message)
                if strict:
                    raise error
                continue
            if isinstance(value, dict) and isinstance(self[key], SettingNode):
                self[key].set_from_dict(value)
            elif type(value) is str:
                self.settings[key] = Setting(
                    self.settings[key].name,
                    self.settings[key].parameter.data_type.cast(value),
                    path=self.settings[key].path,
                )
            else:
                self.settings[key] = self.settings[key].update(value)

    def setting_with_path_name(self, setting: Setting) -> Setting:
        """Get a copy of a setting with its name replaced with the path name."""
        return self.find_by_name(setting.name).with_path_name()

    def diff(self, other: SettingNode, *, path: str = "") -> list[str]:
        """Recursive diff between two SettingNodes.

        This function is meant to produce human-readable output, e.g. for debugging purposes.
        It returns the differences in a list of strings, each string detailing
        one specific difference. The diff is non-symmetric.

        Args:
            other: second node to compare ``self`` to
            path: node path to the currently compared nodes (used in printing the results)

        Returns:
            differences from ``self`` to ``other``, in depth-first order

        """

        def diff_settings(x: Setting, y: Setting, key: str) -> str:
            """Compare two settings, return the differences."""
            line = f"{key}:"

            def compare(x: Any, y: Any, tag: str) -> None:
                """Compare the given properties, add a note to the current line if there is a difference."""
                nonlocal line
                if isinstance(x, np.ndarray):
                    if np.any(x != y):
                        line += f" {tag}: differs"
                    return
                if x != y:
                    line += f" {tag}: {x}/{y}"

            compare(x.name, y.name, "n")
            compare(x.label, y.label, "l")
            compare(x.unit, y.unit, "u")
            compare(x.parameter.data_type, y.parameter.data_type, "dt")
            a_ct = x.parameter.collection_type
            b_ct = y.parameter.collection_type
            compare(a_ct, b_ct, "ct")
            if a_ct == b_ct:
                compare(x.value, y.value, "v")
            return line

        node_diff: list[str] = []
        # compare node names
        if self.name != other.name:
            node_diff.append(f"node name: {self.name}/{other.name}")

        # compare settings
        b_keys = set(other.settings)
        for key, a_setting in self.settings.items():
            b_setting = other.settings.get(key)
            if b_setting is None:
                node_diff.append(f"-setting: {key}")
                continue
            b_keys.remove(key)
            if a_setting != b_setting:
                node_diff.append(diff_settings(a_setting, b_setting, key))
        for key in b_keys:
            if key not in self.settings:
                node_diff.append(f"+setting: {key}")

        # compare subnodes
        diff_subnodes: list[tuple[SettingNode, SettingNode, str]] = []
        b_keys = set(other.subtrees)
        for key, a_sub in self.subtrees.items():
            b_sub = other.subtrees.get(key)
            if b_sub is None:
                node_diff.append(f"-subnode: {key}")
            else:
                b_keys.remove(key)
                diff_subnodes.append((a_sub, b_sub, key))
        for key in b_keys:
            if key not in self.subtrees:
                node_diff.append(f"+subnode: {key}")

        # add path prefixes
        diff = [f"{path}: {d}" for d in node_diff]

        # recurse into subnodes, depth first
        for a_sub, b_sub, key in diff_subnodes:
            diff.extend(a_sub.diff(b_sub, path=f"{path}.{key}" if path else key))

        return diff

    def _withsiprefix(self, val, unit):
        """Turn a numerical value and unit, and return rescaled value and SI prefixed unit.

        Unit must be a whitelisted SI base unit.
        """
        if not isinstance(val, numbers.Real):
            return val, unit
        if unit not in {"Hz", "rad", "s", "V"}:
            return val, unit

        val = float(val)

        pfx = ""
        for p in "kMGP":
            if abs(val) <= 10e3:
                break
            val *= 1e-3
            pfx = p
        for p in "mμnp":
            if not 1 > abs(val) > 0:
                break
            val *= 1e3
            pfx = p

        return val, f"{pfx}{unit}"

    def _repr_html_(self):
        tmpl_path = pathlib.Path(__file__).parent
        jenv = jinja2.Environment(loader=jinja2.FileSystemLoader(tmpl_path), auto_reload=True)

        return jenv.get_template("settingnode_v2.html.jinja2").render(s=self, withsi=self._withsiprefix, startopen=0)

    def get_node_for_path(self, path: str) -> Setting | SettingNode:
        """Return the node corresponding to the given path.

        Args:
            path: The path.

        Returns:
            The node at ``path`` in self.

        Raises:
            ValueError: If the given path cannot be found in self.

        """
        keys = path.split(".")
        node = self
        for index, key in enumerate(keys, start=1):
            if index < len(keys) and key in node.settings:
                raise ValueError(f"Path '{path}' is invalid: '{key}' is a setting, not a node.")
            if key not in node.children:
                raise KeyError(f"Path '{path}' is invalid: key '{key}' is not found in the preceding node {node}.")
            node = node[key]
        return node

    def add_for_path(
        self,
        nodes: Iterable[Setting | Parameter | SettingNode] | dict[str, Setting | Parameter | SettingNode],
        path: str,
        override_values: dict[str, Any] | None = None,
    ) -> None:
        """Add nodes to ``self`` while creating the missing nodes in-between.

        Whether the names and paths are aligned is determined by the attribute ``align_name`` of the current node
        (``self``). All the created missing nodes will use this same ``align_name`` value,
        which determines whether their names will align with their paths.

        Args:
            nodes: Nodes to add as new leaves/branches of ``path``. If of type ``dict``, maps the keys used in
                ``self.settings`` or ``self.subtrees`` to the nodes themselves. If ``align_name=False``, the key and
                the node name can differ, but otherwise the names will be replaced by the path anyways).
            path: Path in ``self`` to which ``nodes`` will be added. If the path or any part (suffix) of it is not
                found in self, the associated nodes will be created automatically.
            override_values: Optionally override the values for the `Settings` corresponding to ``nodes``. This dict
                should have the same structure as ``nodes``, including matching names.

        """
        override_values = override_values or {}
        path_split = path.split(".")
        # find the depth already found in self
        latest_node = self
        levels_to_add = []
        for idx, fragment in enumerate(path_split):
            if fragment in latest_node.children:
                latest_node = latest_node[fragment]
            else:
                levels_to_add = path_split[idx:]
                break
        # add the missing levels
        for fragment in levels_to_add:
            latest_node[fragment] = SettingNode(name=fragment, align_name=latest_node.align_name)
            latest_node = latest_node[fragment]
        # finally add the nodes
        nodes_to_add = nodes.values() if isinstance(nodes, dict) else nodes
        nodes_keys = list(nodes.keys()) if isinstance(nodes, dict) else []
        for idx, node in enumerate(nodes_to_add):
            key = nodes_keys[idx] if isinstance(nodes, dict) else node.name.split(".")[-1]
            if isinstance(node, SettingNode):
                latest_node[key] = node
            else:
                default_value = node.value if isinstance(node, Setting) else None
                parameter = node.parameter if isinstance(node, Setting) else node
                value = override_values.get(node.name) if override_values.get(node.name) is not None else default_value
                latest_node[key] = Setting(parameter, value)

    def get_default_implementation_name(self, gate: str, locus: str | Iterable[str]) -> str:
        """Get the default implementation name for a given gate and locus.

        Takes into account the global default implementation and a possible locus specific implementation and also
        the symmetry properties of the gate.

        NOTE: using this method requires the standard EXA settings tree structure.

        Args:
            gate: The name of the gate.
            locus: Individual qubits, couplers, or combinations.

        Returns:
            The default implementation name.

        """
        gate_settings = self.gate_definitions[gate]
        for impl in gate_settings.children:
            if (
                isinstance(gate_settings[impl], SettingNode)
                and impl not in ("symmetric", "default_implementation")  # sanity check
                and "override_default_for_loci" in gate_settings[impl].children  # backwards compatibility
                and gate_settings[impl].override_default_for_loci.value
            ):
                if isinstance(locus, str):
                    locus = locus.split("__")
                if gate_settings.symmetric.value:
                    loci = list(permutations(locus))
                else:
                    loci = [tuple(locus)]
                for permuted_locus in loci:
                    locus_str = "__".join(permuted_locus)
                    if locus_str in gate_settings[impl].override_default_for_loci.value:
                        return impl
        return gate_settings.default_implementation.value

    def get_gate_node_for_locus(
        self,
        gate: str,
        locus: str | Iterable[str],
        implementation: str | None = None,
    ) -> SettingNode:
        """Get the gate calibration sub-node for the locus given as a parameter if it exists in the settings tree.

        NOTE: using this method requires the standard EXA settings tree structure.

        Args:
            gate: The gate to retrieve the settings for.
            locus: Individual qubits, couplers, or combinations.
            implementation: Using a custom rather than the default gate implementation.

        Returns:
            The settings of the specified locus and gate.

        """
        pulse_settings = self["gates"] if "gates" in self.children else self

        if gate not in pulse_settings.children:
            raise ValueError(f"Gate {gate} cannot be found in the pulse settings.")

        if not implementation:
            implementation = self.get_default_implementation_name(gate, locus)

        if implementation not in pulse_settings[gate].children:
            raise ValueError(f"Gate implementation {implementation} cannot be found in the pulse settings.")

        str_loci = self._get_symmetric_loci(gate, implementation, locus)

        for str_locus in str_loci:
            if str_locus in pulse_settings[gate][implementation].children:
                return pulse_settings[gate][implementation][str_locus]

        raise ValueError(f"Locus {locus} cannot be found in the pulse settings.")

    def get_locus_node_paths_for(self, gate: str, implementations: list[str] | None = None) -> list[str]:
        """Get all the gate locus node paths for a given ``gate``.

        NOTE: using this method requires the standard EXA settings tree structure.

        Args:
            gate: Gate name.
            implementations: optionally limit the paths by these gate implementations.

        Returns:
            The locus node (string) paths corresponding to this gate.

        """
        node_paths = []
        if "gates" not in self.children or gate not in self.gates.children:
            return node_paths
        if implementations is not None:
            impls = [i for i in self.gates[gate].children if i in implementations]
        else:
            impls = self.gates[gate].children
        for impl_name in impls:
            if isinstance(self.gates[gate][impl_name], SettingNode):
                for locus in self.gates[gate][impl_name].children:
                    if isinstance(self.gates[gate][impl_name][locus], SettingNode):
                        node_paths.append(f"{gate}.{impl_name}.{locus}")
        return node_paths

    def get_gate_properties_for_locus(
        self, gate: str, locus: str | Iterable[str], implementation: str | None = None
    ) -> SettingNode:
        """Get the gate characterization sub-node for the locus given as a parameter if it exists in the settings tree.

        NOTE: using this method requires the standard EXA settings tree structure.

        Args:
            gate: The gate to retrieve the settings for.
            locus: Individual qubits, couplers, or combinations.
            implementation: Using a custom rather than the default gate implementation.

        Returns:
            The settings of the specified locus and gate.

        """
        if not implementation:
            implementation = self.get_default_implementation_name(gate, locus)
        gate_properties = self.characterization.gate_properties
        if gate not in gate_properties.children:
            raise ValueError(f"Gate {gate} cannot be found in the characterization settings.")
        if implementation not in gate_properties[gate].children:
            raise ValueError(
                f"Implementation {implementation} of Gate {gate} cannot be found in the characterization settings."
            )

        str_loci = self._get_symmetric_loci(gate, implementation, locus)
        for str_locus in str_loci:
            if str_locus in gate_properties[gate][implementation].children:
                return gate_properties[gate][implementation][str_locus]

        raise ValueError(f"Locus {locus} cannot be found in the gate properties characterization settings.")

    def _get_symmetric_loci(self, gate: str, implementation: str, locus: str) -> list[str]:
        if not isinstance(locus, str):
            if self.gate_definitions[gate][implementation].symmetric.value:
                str_loci = ["__".join(sort_components(locus))]
            elif self.gate_definitions[gate].symmetric.value:
                str_loci = ["__".join(item) for item in list(permutations(locus))]
            else:
                str_loci = ["__".join(locus)]
        else:
            str_loci = [locus]
        return str_loci

    def _get_path(self, key) -> str:
        if not self.path:
            return key
        return f"{self.path}.{key}"



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from typing import Any, Callable

import numpy as np


def get_json_encoder() -> dict[Any, Callable[..., dict[str, Any]]]:
    """Custom JSON encoder for complex number, ndarray or tuple

    Can be used in situation when serialization of JSON can be customised
    (json_encoders config in pydantic models, custom_encoder in jsonable_encoder
    of fastapi, etc.)
    """

    def _encode_complex(obj: complex) -> dict[str, Any]:
        return {"__complex__": "true", "real": obj.real, "imag": obj.imag}

    def _encode_ndarray(obj: np.ndarray) -> dict[str, Any]:
        data_b64 = base64.b64encode(obj.data)
        return {"__ndarray__": "true", "data": data_b64, "dtype": str(obj.dtype), "shape": obj.shape}

    def _encode_tuple(obj: tuple) -> dict[str, Any]:
        return {"__tuple__": "true", "data": obj}

    return {complex: _encode_complex, np.ndarray: _encode_ndarray, tuple: _encode_tuple}


def decode_json(obj: Any) -> Any:
    """Custom json decoder for object, in case it is represented as
    complex number, ndarray or tuple.
    Otherwise decoder won't be applied

    Here is an example of encoded/decoded json with complex number and ndarray:

    .. doctest::

        >>> import base64
        >>> import numpy as np
        >>> from exa.common.helpers.json_helper import decode_json
        >>> complex_encoded = {'__complex__': True, 'real': 3, 'imag': 4}
        >>> decode_json(complex_encoded)
        (3+4j)

        >>> data_b64 = base64.b64encode(np.arange(4).reshape(2, 2))
        >>> ndarray_encoded = {'__ndarray__': True, 'data': data_b64.decode('utf-8'), 'dtype': 'int64', 'shape':[2, 2]}
        >>> decode_json(ndarray_encoded)
        array([[0, 1],
               [2, 3]])
    """
    if isinstance(obj, dict):
        if "__complex__" in obj:
            return complex(obj["real"], obj["imag"])
        elif "__ndarray__" in obj:
            data = base64.b64decode(obj["data"])
            return np.frombuffer(data, obj["dtype"]).reshape(obj["shape"])
        elif "__tuple__" in obj:
            return tuple(obj["data"])
    return obj



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Range specification to define a linearly spaced interval."""

from dataclasses import dataclass
import math

import numpy as np

from exa.common.control.sweep.option.constants import DEFAULT_COUNT
from exa.common.control.sweep.option.sweep_options import SweepOptions
from exa.common.control.sweep.sweep_values import SweepValues


@dataclass(frozen=True)
class StartStopOptions(SweepOptions):
    """Range generation options.

    Values are generated over the interval from `start` to `stop`. For linear sweep range the
    number of generated values can be based either on `count` or `step`. In case `count` is empty
    and `step` is not, `step` is used for calculating `count`. For exponential sweep range only
    `count` is used.
    """

    #: Starting value of interval.
    start: int | float | complex
    #: Stopping value of interval.
    stop: int | float | complex
    #: Number of values to generate. Must be non-negative.
    #: If `count` and `step` are empty, the default value of count is
    #: :const:`exa.common.control.sweep.option.constants.DEFAULT_COUNT`.
    count: int | None = None
    #: Size of spacing between values. Must be non-zero.
    #: If both `count` and `step` are not empty, only `count` is used
    step: int | float | complex | None = None

    def __post_init__(self):
        if self.count is not None and self.step is not None:
            object.__setattr__(self, "step", None)
        if self.count is None and self.step is None:
            object.__setattr__(self, "count", DEFAULT_COUNT)
        if self.step is not None and (abs(self.step) > abs(self.stop - self.start) or self.step == 0):
            raise ValueError(
                "Step value specified for range must be less than absolute difference of stop and start values and"
                "must be greater than zero."
            )
        if self.count is not None and self.count <= 0:
            raise ValueError("Count value specified for range must be greater than zero.")

    @property
    def data(self) -> list[int | float | complex]:
        if self.step is not None:
            count = 1 + math.ceil(abs(self.stop - self.start) / float(np.abs(self.step)))
            data = self._generate_by_count(count)
        else:
            data = self._generate_by_count(self.count)
        return data

    def _generate_by_count(self, count: int) -> SweepValues:
        return np.linspace(self.start, self.stop, count, endpoint=True).tolist()



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import xarray as xr

"""Helper methods for data manipulation.
"""


def add_data_array(ds: xr.Dataset, da: xr.DataArray, name: Optional[str] = None) -> xr.Dataset:
    """Add data array `da` to dataset `ds`.

    Unlike the default xarray command, preserves metadata of the dataset.

    Args:
        ds: Dataset to add to.
        da: DataArray to add
        name: name under which `da` can be accessed inside `ds`.
            By default, uses the `name` property of `da`.

    Returns:
        The updated dataset.

    """
    if name is None:
        if da.name is not None:
            name = da.name
        else:
            raise ValueError("No name was given to the dataArray.")
    # Attributes of Dataset coordinates are dropped/replaced when adding a DataArray
    # https://github.com/pydata/xarray/issues/2245
    # So we need to temporarily store all coord attrs, and then add them back
    attributes = {}
    for key in ds.coords:
        attributes[key] = ds.coords[key].attrs
    for key in ds.data_vars:
        attributes[key] = ds.data_vars[key].attrs
    ds[name] = da
    for key in ds.coords:
        if attributes.get(key):
            ds.coords[key].attrs = attributes.get(key)
    for key in ds.data_vars:
        if attributes.get(key):
            ds.data_vars[key].attrs = attributes[key]
    return ds



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convert NdSweeps to protos and back."""

import iqm.data_definitions.common.v1.sweep_pb2 as spb

from exa.common.api.proto_serialization import sequence
import exa.common.api.proto_serialization._parameter as param_proto
from exa.common.control.sweep.sweep import Sweep
from exa.common.data.parameter import DataType, Parameter
from exa.common.sweep.util import NdSweep


def pack(nd_sweep: NdSweep, minimal: bool = True) -> spb.CartesianSweep:
    """Convert an NdSweep into protobuf representation.

    Note: The protobuf does not make any distinction between different types of Sweeps, so the type information is lost.
    """
    parallels = []
    for parallel in nd_sweep:
        parallel_proto = spb.ParallelSweep()
        parallel_proto.single_parameter_sweeps.MergeFrom((_pack_single_sweep(sweep, minimal) for sweep in parallel))
        parallels.append(parallel_proto)
    proto = spb.CartesianSweep()
    proto.parallel_sweeps.MergeFrom(reversed(parallels))  # In data-definitions, order is outermost loop first
    return proto


def unpack(proto: spb.CartesianSweep) -> NdSweep:
    """Convert protobuf representation into a NdSweep. Reverse operation of :func:`.pack`."""
    nd_sweep = []
    for parallel_proto in reversed(proto.parallel_sweeps):
        parallel = tuple(_unpack_single_sweep(sweep) for sweep in parallel_proto.single_parameter_sweeps)
        nd_sweep.append(parallel)
    return nd_sweep


def _pack_single_sweep(sweep: Sweep, minimal: bool) -> spb.SingleParameterSweep:
    kwargs = {"parameter_name": sweep.parameter.name, "values": sequence.pack(sweep.data)}
    if not minimal:
        kwargs["parameter"] = param_proto.pack(sweep.parameter)
    return spb.SingleParameterSweep(**kwargs)


def _unpack_single_sweep(proto: spb.SingleParameterSweep) -> Sweep:
    sweep_values = sequence.unpack(proto.values)
    if proto.HasField("parameter"):
        parameter = param_proto.unpack(proto.parameter)
    else:
        parameter = Parameter(proto.parameter_name, data_type=DataType.ANYTHING)
    return Sweep(parameter=parameter, data=sweep_values)



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sweep specification with exponentially spaced values."""

from typing import Any
import warnings

from exa.common.control.sweep.option import CenterSpanBaseOptions, StartStopBaseOptions
from exa.common.control.sweep.sweep import Sweep
from exa.common.data.parameter import Parameter
from exa.common.errors.exa_error import InvalidSweepOptionsTypeError


class ExponentialSweep(Sweep):
    """Generates parameter values spaced evenly on a geometric progression based on `options`.

    - If `options` is instance of :class:`.StartStopBaseOptions`,
      the start and stop of the interval are calculated from powers of start and stop.
    - If `options` is instance of :class:`.CenterSpanBaseOptions`,
      the start and stop of the interval are calculated from powers of start and stop,
      which are derived from center and span.

    Raises:
        ValueError: Error is raised if `options` is inconsistent.

    """

    def __init__(
        self,
        parameter: Parameter,
        options: StartStopBaseOptions | CenterSpanBaseOptions | None = None,
        *,
        data: list[Any] | None = None,
        **kwargs,
    ) -> None:
        warnings.warn("ExponentialSweep is deprecated, use Sweep instead.", DeprecationWarning)
        if options and not isinstance(options, StartStopBaseOptions | CenterSpanBaseOptions):
            raise InvalidSweepOptionsTypeError(str(type(options)))
        super().__init__(parameter, options, data=data, **kwargs)



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sweep specification with arbitrary values."""

from typing import Any
import warnings

from exa.common.control.sweep.option import FixedOptions
from exa.common.control.sweep.sweep import Sweep
from exa.common.data.parameter import Parameter
from exa.common.errors.exa_error import InvalidSweepOptionsTypeError


class FixedSweep(Sweep):
    """A sweep over arbitrary set of values, given by `options`."""

    def __init__(
        self, parameter: Parameter, options: FixedOptions | None = None, *, data: list[Any] | None = None, **kwargs
    ) -> None:
        warnings.warn("FixedSweep is deprecated, use Sweep instead.", DeprecationWarning)
        if options and not isinstance(options, FixedOptions):
            raise InvalidSweepOptionsTypeError(str(type(options)))
        super().__init__(parameter, options, data=data, **kwargs)



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Range specification used with exponential sweeps."""

from dataclasses import dataclass
import logging
import math

import numpy as np

from exa.common.control.sweep.option.constants import DEFAULT_BASE, DEFAULT_COUNT
from exa.common.control.sweep.option.sweep_options import SweepOptions

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class StartStopBaseOptions(SweepOptions):
    """Range generation options.

    Values are generated over the interval from `base` power `start` to `base` power `stop`.
    The number of values = `count`. These options are used only for exponential sweep range.
    """

    #: The power for the start of the interval.
    start: int | float | complex
    #: The power for the end of the interval.
    stop: int | float | complex
    #: Number of values to generate. Default to
    #: :const:`exa.common.control.sweep.option.constants.DEFAULT_COUNT`.
    count: int | None = None
    #: Number, that is raised to the power `start` or `stop`. Default to
    #: :const:`exa.common.control.sweep.option.constants.DEFAULT_BASE`.
    base: int | float | None = None

    def __post_init__(self):
        if self.count is None:
            object.__setattr__(self, "count", DEFAULT_COUNT)
        if self.base is None:
            object.__setattr__(self, "base", DEFAULT_BASE)
        if self.start == 0 or self.stop == 0:
            raise ValueError("Exponential range sweep start and stop values must not be zero.")

    @property
    def data(self) -> list[int | float | complex]:
        logger.debug(f"EXPONENTS: ({self.start}, {self.stop}) with base {self.base}")
        start = math.pow(self.base, self.start)
        stop = math.pow(self.base, self.stop)
        return np.geomspace(start, stop, self.count, endpoint=True).tolist()



class ExaError(Exception):
    """Base class for exa errors.

    Attributes:
        message: Error message.

    """

    def __init__(self, message: str, *args):
        super().__init__(message, *args)
        self.message = message


class UnknownSettingError(ExaError, AttributeError):
    """This SettingNode does not have a given key."""


class InvalidParameterValueError(ExaError, ValueError):
    """The value set does not conform to the parameter restrictions."""


class InvalidSweepOptionsTypeError(ExaError, TypeError):
    """The type of sweep options is invalid."""

    def __init__(self, options: str, *args):
        super().__init__(f"Options have unsupported type of {options}", *args)


class RequestError(ExaError):
    """Error raised when something went wrong on the server after sending a request."""


class EmptyComponentListError(ExaError, ValueError):
    """Error raised when an empty list is given as components for running an experiment."""



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pack and unpack parameters to protos and back."""

import iqm.data_definitions.common.v1.parameter_pb2 as ppb

from exa.common.data.parameter import CollectionType, DataType, Parameter

_COLLECTION_TYPES = {
    CollectionType.SCALAR: ppb.Parameter.CollectionType.COLLECTION_TYPE_SCALAR,
    CollectionType.LIST: ppb.Parameter.CollectionType.COLLECTION_TYPE_SEQUENCE,
    CollectionType.NDARRAY: ppb.Parameter.CollectionType.COLLECTION_TYPE_ARRAY,
}
_collection_types_inv = {v: k for k, v in _COLLECTION_TYPES.items()}

_DATA_TYPES = {
    DataType.ANYTHING: ppb.Parameter.DataType.DATA_TYPE_UNSPECIFIED,
    DataType.NUMBER: ppb.Parameter.DataType.DATA_TYPE_FLOAT64,  # Deprecated
    DataType.STRING: ppb.Parameter.DataType.DATA_TYPE_STRING,
    DataType.COMPLEX: ppb.Parameter.DataType.DATA_TYPE_COMPLEX128,
    DataType.BOOLEAN: ppb.Parameter.DataType.DATA_TYPE_BOOL,
    DataType.INT: ppb.Parameter.DataType.DATA_TYPE_INT64,
    DataType.FLOAT: ppb.Parameter.DataType.DATA_TYPE_FLOAT64,
}
_data_types_inv = {v: k for k, v in _DATA_TYPES.items() if k != DataType.NUMBER}


def pack(parameter: Parameter) -> ppb.Parameter:
    """Convert Parameter into protobuf representation."""
    return ppb.Parameter(
        name=parameter.name,
        label=parameter.label,
        unit=parameter.unit,
        data_type=_DATA_TYPES.get(parameter.data_type, ppb.Parameter.DataType.DATA_TYPE_UNSPECIFIED),
        collection_type=_COLLECTION_TYPES[parameter.collection_type],
        element_indices=parameter.element_indices,
        parent_name=parameter.parent_name,
        parent_label=parameter.parent_label,
    )


def unpack(proto: ppb.Parameter) -> Parameter:
    """Convert protobuf representation into a Parameter."""
    if not proto.element_indices:
        return Parameter(
            name=proto.name,
            label=proto.label,
            unit=proto.unit,
            data_type=_data_types_inv.get(proto.data_type),
            collection_type=_collection_types_inv.get(proto.collection_type),
        )
    return Parameter(
        name=proto.parent_name,
        label=proto.parent_label,
        unit=proto.unit,
        data_type=_data_types_inv.get(proto.data_type),
        collection_type=_collection_types_inv.get(proto.collection_type),
        element_indices=list(proto.element_indices),
    )



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions that encode and decode all sweep arguments(different types of
sweeps, return parameters, settings, etc.). The module provides functions for
serializing and deserializing sweep arguments before saving them to database.
"""

import json
from typing import Any, cast

from exa.common.control.sweep.sweep import Sweep
from exa.common.data.parameter import Parameter
from exa.common.data.setting_node import SettingNode
from exa.common.helpers import json_helper
from exa.common.helpers.json_helper import decode_json, get_json_encoder
from exa.common.sweep.util import NdSweep, Sweeps


def encode_nd_sweeps(sweeps: NdSweep, **kwargs) -> str:
    """Encode sweeps to a JSON string.

    Args:
        sweeps: sweeps to be serialized.
        kwargs: keyword arguments passed to json.dumps

    Returns:
        json as a string

    """
    kwargs.setdefault("cls", _SweepDataEncoder)
    return json.dumps(sweeps, **kwargs)


def encode_return_parameters(return_parameters: dict[Parameter, NdSweep | None], **kwargs) -> str:
    """Encode sweeps to a JSON string.

    Args:
        return_parameters: Return parameters as specified by :meth:`~.Backend.sweep`.
        kwargs: keyword arguments passed to json.dumps

    Returns:
        json as a string

    """
    kwargs.setdefault("cls", _SweepDataEncoder)
    listed_params = [{"parameter": param, "hard_sweeps": sweeps} for param, sweeps in return_parameters.items()]
    return json.dumps(listed_params, **kwargs)


def decode_and_validate_sweeps(sweeps_json: str) -> Sweeps:
    """Decodes and validates json string of list of Sweeps and tuples of Sweeps.

    Args:
        sweeps_json: json string of nested structure of Sweep objects

    Returns:
        list of sweeps and tuples of sweeps

    Raises:
        ValueError if decoded result is not expected return type

    """
    decoded = _loads(sweeps_json)
    if not isinstance(decoded, list):
        raise ValueError(f"Outer type is not list type when decoding {sweeps_json}")

    for elem in decoded:
        if isinstance(elem, tuple):
            if any([not isinstance(t, Sweep) for t in elem]):
                raise ValueError("list of Sweeps must contain tuples of `Sweep` objects")
        elif not isinstance(elem, Sweep):
            raise ValueError(f"list must contain either Sweep or Tuple elements. Got {type(elem)}")
    return cast(Sweeps, decoded)


def decode_return_parameters(json_str: str) -> dict[Parameter, NdSweep | None]:
    """Deserialize return parameters.

    For backwards compatibility, changes values of the return parameters dict to a new,
    more general format: NdSweeps, which is a list of tuples of Sweeps.

    Return parameters JSON syntax: ``[{"parameter": readout_parameter_json, "hard_sweeps": [hard_sweep,...]}, ...]``.

    ``readout_parameter_json`` may be a Parameter or a Setting; both will be converted to a Parameter.
    ``hard_sweeps`` are converted with the following rules:

    * 1 is converted to an empty NdSweep, i.e., a scalar.
    * Other integers are converted to a :func:`.linear_index_sweep`
    * `Sweep` is converted to an NdSweep that contains only the sweep
    * NdSweep and None are not converted.

    Args:
        json_str: JSON representation of the ``return_parameters`` loaded
            from e.g. persistence

    Return:
        a reconstituted, typed ``return_parameters`` structure

    """
    decoded = _loads(json_str)
    if not isinstance(decoded, list):
        raise ValueError(f"Outer type is not list type when decoding: {json_str}")

    return_parameters = {
        Parameter(**param_and_sweeps["parameter"]): param_and_sweeps["hard_sweeps"] for param_and_sweeps in decoded
    }
    return return_parameters


def decode_settings(json_str: str) -> SettingNode:
    """Deserialize settings from json string

    Args:
        json_str: JSON representation of ``settings`` loaded
            from e.g. persistence

    Returns:
        deserialized settings

    """
    json_loads = json.loads(json_str)
    return SettingNode(**json_loads)


def _loads(*args, **kwargs) -> Any:
    """Extend json.loads with tuple and Sweep decoding support

    Args:
        args: positional arguments passed to json.dumps
        kwargs: keyword arguments passed to json.dumps

    Returns:
        python data structure that recovers encoded structure of
        nested tuples and Sweep -objects

    """

    def _decode_json(obj: Any) -> Any:
        if isinstance(obj, dict) and "parameter" in obj:
            # This is kind of a hack to provide backwards compatibility for old data.
            #  Pydantic doesn't allow extra attributes with the current setup,
            #  but old database data can have "parent_name" and "parent_label", which are not part of the model.
            obj["parameter"].pop("parent_name", None)
            obj["parameter"].pop("parent_label", None)

        if isinstance(obj, dict) and {"parameter", "data"}.issubset(obj):
            data = [json_helper.decode_json(d) for d in obj["data"]]
            return Sweep(parameter=Parameter(**obj["parameter"]), data=data)
        json_data_return = decode_json(obj)
        return json_data_return

    kwargs.setdefault("object_hook", _decode_json)
    loads = json.loads(*args, **kwargs)
    return loads


class _SweepDataEncoder(json.JSONEncoder):
    """Extension of json encoder to support handling of Sweep, Parameter, and tuples."""

    def default(self, obj: Any) -> Any:
        """Default encoder

        Args:
            obj: object to be encoded

        Returns:
            encoded object

        """
        if isinstance(obj, Sweep):
            obj = obj.model_copy().model_dump()
        elif isinstance(obj, Parameter):
            obj = obj.model_copy().model_dump()
        elif isinstance(obj, tuple):
            obj = get_json_encoder()[tuple](obj)
        else:
            obj = super().default(obj)
        return obj

    # JSONEncoder doesn't call `default` method for tuple, because
    # it knows how to serialize it. In order to use custom encoder
    # we need to override `encode` method
    def encode(self, obj: Any) -> Any:
        """Override JSONEncoder encode in order to change the
        default tuple behaviour.

        Args:
            obj: object to be encoded

        Returns:
            encoded object

        """

        def _encode_tuples(item):
            if isinstance(item, tuple):
                return get_json_encoder()[tuple](item)
            if isinstance(item, list):
                return [_encode_tuples(e) for e in item]
            if isinstance(item, dict):
                return {
                    key if isinstance(key, str) else self.encode(key): _encode_tuples(value)
                    for key, value in item.items()
                }
            return item

        return super().encode(_encode_tuples(obj))



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import cached_property

from pydantic import BaseModel, ConfigDict


# Change the behaviour of pydantic globally
class ImmutableBaseModel(BaseModel):
    # Unable to use cached_property
    # https://github.com/pydantic/pydantic/issues/1241
    model_config = ConfigDict(frozen=True, ignored_types=(cached_property,))



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic utilities for converting sweep definitions from
user-friendly format to canonic ones.
"""

from exa.common.control.sweep.option import StartStopOptions
from exa.common.control.sweep.sweep import Sweep
from exa.common.data.parameter import Parameter

# One sweep where many sweeps are executed together, like a zip. Type hint alias.
ParallelSweep = tuple[Sweep, ...]

# N-dimensional sweep, represented as list of N parallel Sweeps. Type hint alias.
NdSweep = list[ParallelSweep]

# List of sweeps or parallel Sweeps. Type hint alias.
Sweeps = list[Sweep | ParallelSweep]


def convert_sweeps_to_list_of_tuples(sweeps: Sweeps) -> NdSweep:
    """Validate sweeps and convert it to format accepted by the station control.

    Converts a more convenient sweep definition list to a strict list of tuples of sweeps.
    The sweep instances themselves are the same, except single sweep instances are turned
    into a tuple containing a single sweep.

    Verify that:
    * sweeps list element is either Sweep or ParallelSweep
    * tuple_of_sweep element is a Sweep
    * tuple of sweeps contains at least one element
    * length of a data is identical in all tuples of sweeps

    Args:
        sweeps: More user-friendly definition of a list of sweeps.

    Returns:
        List of tuples of sweeps.

    Raises:
        ValueError if sweeps parameter does not follow the contract.

    """
    new_list = []
    for tuple_or_sweep in sweeps:
        if isinstance(tuple_or_sweep, tuple):
            if not len(tuple_or_sweep) > 0:
                raise ValueError("Tuples of sweeps must have at least one element")
            for sweep in tuple_or_sweep:
                if not isinstance(sweep, Sweep):
                    raise ValueError(f"Tuples of sweeps must contain tuple type, got {type(sweep)}")
                if len(sweep.data) != len(tuple_or_sweep[0].data):
                    raise ValueError(
                        f"Length {len(sweep.data)} of a data in {sweep} did not match to expected "
                        + f" length {len(tuple_or_sweep[0].data)}"
                    )
            new_list.append(tuple_or_sweep)
        elif isinstance(tuple_or_sweep, Sweep):
            new_list.append((tuple_or_sweep,))
        else:
            raise ValueError(f"Elements in sweeps must be either tuples of Sweeps or Sweep, got {type(tuple_or_sweep)}")
    return new_list


def linear_index_sweep(parameter: Parameter, length: int) -> list[tuple[Sweep]]:
    """Produce an NdSweep over a dummy index.

    Can be used in places where a "hardware sweep" is needed but not really meaningful.

    Args:
        parameter: Data parameter this index is for.
        length: Number of integers in the dummy sweep.

    Returns:
        A linear sweep over a parameter whose name is ``parameter.name + _index`` and whose data ranges from 0 to
        `length` with steps of 1.

    """
    return [
        (
            Sweep(
                parameter=Parameter(name=parameter.name + "_index", label=parameter.label + " index"),
                data=StartStopOptions(0, length - 1, count=length).data,
            ),
        )
    ]



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base immutable class for sweeps specifications."""

from typing import Any
import warnings

from exa.common.control.sweep.option import CenterSpanOptions, StartStopOptions, SweepOptions
from exa.common.control.sweep.sweep_values import SweepValues
from exa.common.data.base_model import BaseModel
from exa.common.data.parameter import Parameter
from exa.common.errors.exa_error import InvalidSweepOptionsTypeError


class Sweep(BaseModel):
    """Base immutable class for sweeps."""

    parameter: Parameter
    """The Sweep represents changing the values of this Parameter."""

    data: SweepValues
    """List of values for :attr:`parameter`"""

    def __init__(
        self, parameter: Parameter, options: SweepOptions | None = None, *, data: SweepValues | None = None, **kwargs
    ) -> None:
        if options is None and data is None:
            raise ValueError("Either 'options' or 'data' is required.")
        if options is not None and data is not None:
            raise ValueError(
                "Can't use both 'options' and 'data' at the same time, give only either of the parameters."
            )
        if options is not None:
            warnings.warn("'options' attribute is deprecated, use 'data' instead.", DeprecationWarning)

            if not isinstance(options, SweepOptions):
                raise InvalidSweepOptionsTypeError(str(type(options)))

            if isinstance(options, StartStopOptions):
                data = self.__from_start_stop(parameter, options)
            elif isinstance(options, CenterSpanOptions):
                data = self.__from_center_span(parameter, options)
            else:
                data = options.data

        super().__init__(parameter=parameter, data=data, **kwargs)

    def model_post_init(self, __context: Any) -> None:
        if not all(self.parameter.validate(value) for value in self.data):
            raise ValueError(f"Invalid range data {self.data} for parameter type {self.parameter.data_type}.")

    @classmethod
    def __from_center_span(cls, parameter: Parameter, options: CenterSpanOptions) -> SweepValues:
        cls._validate_value(parameter, options.center, "center")
        cls._validate_value(parameter, options.span, "span")
        return options.data

    @classmethod
    def __from_start_stop(cls, parameter: Parameter, options: StartStopOptions) -> SweepValues:
        cls._validate_value(parameter, options.start, "start")
        cls._validate_value(parameter, options.stop, "stop")
        cls._validate_value(parameter, options.step, "step")
        return options.data

    @staticmethod
    def _validate_value(parameter: Parameter, value: int | float | complex | str | bool, value_label: str):
        if not parameter.validate(value):
            raise ValueError(f"Invalid {value_label} value {value} for parameter type {parameter.data_type}.")



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Coercion of numpy types."""

from typing import Any

import numpy as np


def coerce_numpy_type_to_native(value: Any) -> Any:
    """Convert numpy types to underlying native types, and Windows-specific int32 arrays to int64s."""
    if isinstance(value, np.ndarray) and value.dtype == np.int32:
        return value.astype(dtype=np.int64, casting="same_kind")

    if isinstance(value, (np.number, np.bool_)):
        return value.item()  # return the closest native type.

    return value



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convert Python Sequences to protos and back."""

from typing import Sequence

import iqm.data_definitions.common.v1.data_types_pb2 as dpb
import numpy as np


def pack(values: Sequence) -> dpb.Sequence:
    """Packs a sequence of native Python types into protobuf format.

    Args:
        values: Sequence to convert.

    Returns:
        A protobuf instance that encapsulates `values`.

    Raises:
        ValueError in case of unsupported value.

    """
    target = dpb.Sequence()
    if not values:
        return target
    dtype = type(values[0])
    if dtype is complex or np.issubdtype(dtype, np.complexfloating):  # dtype is complex check for contingency
        target_field = target.complex128_array
        target_field.real.MergeFrom(np.real(values))
        target_field.imag.MergeFrom(np.imag(values))
        return target
    if dtype is bool:
        target_field = target.bool_array
    elif dtype is int or np.issubdtype(dtype, np.integer):
        target_field = target.int64_array
    elif dtype is float or np.issubdtype(dtype, np.floating):
        target_field = target.float64_array
    elif dtype is str:
        target_field = target.string_array
    else:
        raise TypeError(f"Unsupported numpy array type {dtype} for a sequence.")

    target_field.items.MergeFrom(values)
    return target


def unpack(source: dpb.Sequence) -> list:
    """Unpacks protobuf to list. Reverse operation of :func:`.pack`.

    Args:
        source: A protobuf instance that encapsulates some data.

    Returns:
        Unpacked data.

    Raises:
        ValueError or google.protobuf.message.DecodeError in case of invalid buffer

    """
    kind = source.WhichOneof("kind")
    if kind is None:
        return []

    if kind == "complex128_array":
        return list(
            complex(real, imag) for real, imag in zip(source.complex128_array.real, source.complex128_array.imag)
        )
    return list(getattr(source, kind).items)



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from importlib import reload
from importlib.metadata import distribution
import os
import platform
import subprocess
from typing import Optional

import pkg_resources


def _is_editable(pkg_name: str) -> bool:
    """Determine if the package is installed in the editable/development mode.

    Currently happens by checking if the first file in the installed distribution has the name
    __editable__.<pkgname>-<version>.pth.
    TODO this is a hacky way of figuring out editability, not an official feature of ``setuptools`` or
    ``importlib.metadata``, so it might break anytime.
    """
    dist = distribution(pkg_name)
    return dist.files and dist.files[0].name.startswith("__editable__.")


def get_all_software_versions(reload_module: bool = False) -> dict[str, str]:
    """Get all available software version information.

    Currently, this function collects all Python package versions and Python interpreter version.

    Args:
        reload_module: Whether to reload the ``pkg_resources`` module or not. By default,
            it is disabled because reloading the module is not thread safe!
            This function should be called with ``reload_module=True`` when IPython autoreload is in use.

    Example:
            1. You have numpy==1.21.0 installed, and in the notebook you have executed the following IPython magic:

                .. code-block:: python

                    %load_ext autoreload
                    %autoreload 2

            2. You install numpy==1.21.1

            3. You call this function with ``reload_module=False``. This will result in some warning printouts and
               the function will succeed by returning 1.21.0 for numpy, which is wrong because in reality IPython
               autoreload has reloaded the newly installed numpy 1.21.1.
               With ``reload_module=True`` the correct version 1.21.1 is returned and no warnings are printed.


    Returns: All software components in a dictionary that maps each package name to its version
        information. A package's version information contains the base version, and the string
        "(local editable)" in the case the package is a local editable installation.

    """
    python_version = platform.python_version()
    software_versions = {"python": python_version}
    # TODO use of pkg_resources is discouraged, replace it with importlib.metadata below.
    # https://setuptools.pypa.io/en/latest/pkg_resources.html
    if reload_module:
        reload(pkg_resources)
    for pkg in pkg_resources.working_set:
        value = pkg.parsed_version.base_version
        if _is_editable(pkg.project_name):
            value += " (local editable)"
        software_versions[pkg.project_name] = value
    return software_versions


def get_vcs_description(root_directory: str) -> Optional[str]:
    """Get Version Control System (VCS) description for the caller's current working directory.

    The description is used to verify if a directory is installed under VCS and whether changes to the files have
    been made. First, the existence of ``.git`` directory will be checked from :attr:`root_directory`.
    Only if it exists, a ``git`` command is executed in a subprocess with a timeout
    of 1 seconds as the best effort only.

    Attributes:
        root_directory: The path to the directory where the command will be executed. For instance when called from
            exa-experiment, it can be the exa-experiment root directory or any directory under it.

    Returns:
        If :attr:`root_directory` is not installed under git, None will be returned. Otherwise, the output of
        ``git describe --dirty --tags --long`` is returned. In case of errors in executing the command, the caught
        ``subprocess.CalledProcessError`` will be converted to string and returned.

    Raises:
        If the command fails or timeouts, an exception will be raised directly from ``subprocess.check_output``.

    """
    # TODO does not seem very robust, consider using the full version number (e.g. 17.0.post1.dev9+g4748363.d20221003)
    # provided by setuptools_scm and accessed using importlib.metadata.version instead, if
    # it can be made to work with importlib.reload.
    if not os.path.isdir(os.path.join(root_directory, ".git")):
        return None

    # git command is similar to what setuptools_scm uses for describing versions:
    # SETUPTOOLS_SCM_DEBUG=1 python3 setup.py --version
    return subprocess.check_output(
        ["git", "describe", "--dirty", "--tags", "--long"], timeout=5, text=True, cwd=root_directory
    ).strip()
